@@ -26,16 +26,6 @@
 /* Find a node in the node list.				*/
 /*--------------------------------------------------------------*/
 
-/* Define record holding information pointing to a gate and the	*/
-/* index into a specific node of that gate.			*/
-
-typedef struct gatenode_ *GATENODE;
-
-struct gatenode_ {
-    GATE gate;
-    int idx;
-};
-
 GATE
 FindGateNode(Tcl_HashTable *NodeTable, NODE node, int *ridx)
 {
@@ -62,13 +52,11 @@ typedef struct _endpointinfo {
     ROUTE orig;			/* original pointer to routed segments */
     int startx;			/* values at segment start */
     int starty;
-    int startl;
-    u_char starttype;
+    int startl;			/* note: layer is exact, not base via layer */
     NODE startnode;
     int endx;			/* values at segment end */
     int endy;
     int endl;
-    u_char endtype;
     NODE endnode;
     double res;			/* total resistance of segment	*/
     double cap;			/* total capacitance of segment */
@@ -143,39 +131,17 @@ check_downstream(SEG walkseg, endpointinfo *eptinfo, int eidx,
 
 	/* Check wire/via layer compatibility */
 
-	if (eptinfo[i].starttype & ST_WIRE) {
-	    if (walkseg->segtype & ST_WIRE)
-		startcompat = (walkseg->layer == eptinfo[i].startl);
-	    else
-		startcompat = (walkseg->layer == eptinfo[i].startl) ||
+	if (walkseg->segtype & ST_WIRE)
+	    startcompat = (walkseg->layer == eptinfo[i].startl);
+	else
+	    startcompat = (walkseg->layer == eptinfo[i].startl) ||
 			(walkseg->layer + 1 == eptinfo[i].startl);
-	}
-	else {
-	    if (walkseg->segtype & ST_WIRE)
-		startcompat = (walkseg->layer == eptinfo[i].startl) ||
-			(walkseg->layer == eptinfo[i].startl + 1);
-	    else
-		startcompat = (walkseg->layer == eptinfo[i].startl) ||
-			(walkseg->layer == eptinfo[i].startl + 1) ||
-			(walkseg->layer + 1 == eptinfo[i].startl);
-	}
 
-	if (eptinfo[i].endtype & ST_WIRE) {
-	    if (walkseg->segtype & ST_WIRE)
-		endcompat = (walkseg->layer == eptinfo[i].endl);
-	    else
-		endcompat = (walkseg->layer == eptinfo[i].endl) ||
+	if (walkseg->segtype & ST_WIRE)
+	    endcompat = (walkseg->layer == eptinfo[i].endl);
+	else
+	    endcompat = (walkseg->layer == eptinfo[i].endl) ||
 			(walkseg->layer + 1 == eptinfo[i].endl);
-	}
-	else {
-	    if (walkseg->segtype & ST_WIRE)
-		endcompat = (walkseg->layer == eptinfo[i].endl) ||
-			(walkseg->layer == eptinfo[i].endl + 1);
-	    else
-		endcompat = (walkseg->layer == eptinfo[i].endl) ||
-			(walkseg->layer == eptinfo[i].endl + 1) ||
-			(walkseg->layer + 1 == eptinfo[i].endl);
-	}
 
 	if ((walkseg->x2 == eptinfo[i].startx) &&
 		(walkseg->y2 == eptinfo[i].starty) && startcompat) {
@@ -258,7 +224,6 @@ walk_route(int eidx, int driverend, endpointinfo *eptinfo,
     GATE g;
     NODE node;
     int i;
-    u_char f;
     ROUTE rt;
 
     eptinfo[eidx].flags |= EPT_VISITED;
@@ -312,10 +277,6 @@ walk_route(int eidx, int driverend, endpointinfo *eptinfo,
 	i = eptinfo[eidx].startl;
 	eptinfo[eidx].startl = eptinfo[eidx].endl;
 	eptinfo[eidx].endl = i;
-
-	f = eptinfo[eidx].starttype;
-	eptinfo[eidx].starttype = eptinfo[eidx].endtype;
-	eptinfo[eidx].endtype = f;
     }
     else
 	firstseg = rt->segments;
@@ -396,7 +357,10 @@ walk_route_output(endpointinfo *eptinfo, int eidx,
 	    fprintf(delayFile, "%s/%s ", g->gatename, g->gatetype->node[i]);
 	if (d > 0) fprintf(delayFile, ", ");
     }
-
+    else if (d == 0) {
+	/* This should not happen:  No node, no route */
+	fprintf(delayFile, "ERROR ");
+    }
 
     /* Output downstream nodes */
     for (i = 0; i < d; i++) {
@@ -409,6 +373,24 @@ walk_route_output(endpointinfo *eptinfo, int eidx,
     fprintf(delayFile, ") ");
 }
 
+/*--------------------------------------------------------------*/
+/* Free data associated with each entry in the NodeTable hash.	*/
+/*--------------------------------------------------------------*/
+
+int FreeNodeTable(Tcl_HashTable *NodeTable)
+{
+    Tcl_HashEntry *entry;
+    Tcl_HashSearch hs;
+    GATENODE gn;
+
+    entry = Tcl_FirstHashEntry(NodeTable, &hs);
+    while (entry != NULL) {
+	gn = Tcl_GetHashValue(entry);
+	if (gn != NULL) free(gn);
+	entry = Tcl_NextHashEntry(&hs);
+    }
+}
+
 /*--------------------------------------------------------------*/
 /* Write an output file of the calculated R, C for every route	*/
 /* branch.  Because the qrouter algorithm is agnostic about the	*/
@@ -431,7 +413,7 @@ int write_delays(char *filename)
     NODEINFO nodeptr;
     SEG seg, newseg, lastseg, nxseg;
     GATE g, drivergate;
-    int i, j, n, new, driverend;
+    int i, j, n, new, driverend, testl;
     int drivernodeidx, driveridx;
     int nroute, numroutes;
     endpointinfo *eptinfo;
@@ -479,11 +461,11 @@ int write_delays(char *filename)
 	LefGetRouteRCvalues(i, &areacap, &edgecap, &respersq);
 	width = LefGetRouteWidth(i);
 
-	lefrcvalues[i].resx = (PitchX[i] / width) * respersq;
-	lefrcvalues[i].resy = (PitchY[i] / width) * respersq;
+	lefrcvalues[i].resx = (PitchX / width) * respersq;
+	lefrcvalues[i].resy = (PitchY / width) * respersq;
 
-	lefrcvalues[i].capx = (PitchX[i] * width) * areacap + (PitchX[i] * edgecap);
-	lefrcvalues[i].capy = (PitchY[i] * width) * areacap + (PitchY[i] * edgecap);
+	lefrcvalues[i].capx = (PitchX * width) * areacap + (PitchX * edgecap);
+	lefrcvalues[i].capy = (PitchY * width) * areacap + (PitchY * edgecap);
 
 	if (i < (Num_layers - 1))
 	    LefGetViaResistance(i, &(lefrcvalues[i].viares));
@@ -496,16 +478,14 @@ int write_delays(char *filename)
     for (n = 0; n < Numnets; n++) {
 	net = Nlnets[n];
 
-	if ((net->netnum == VDD_NET) || (net->netnum == GND_NET)) continue;
+	if ((net->netnum == VDD_NET) || (net->netnum == GND_NET) ||
+		(net->netnum == ANTENNA_NET)) continue;
 
 	/* Count number of net routes */
 	numroutes = 0;
 	for (rt = net->routes; rt; rt = rt->next) numroutes++;
 	if (numroutes == 0) continue;	/* Ignore nets with no routes */
 
-	/* Marked as one driver node.  Not handling more than one driver yet. */
-        fprintf(delayFile, "%s 1", net->netname);
-
 	/* Determine the driver node, as determined by the node with	*/
 	/* LEF direction 'OUTPUT'.  					*/
 	/* (For now, if a net has multiple tristate drivers, just use	*/
@@ -530,18 +510,31 @@ int write_delays(char *filename)
 		eptinfo[nroute].startx = seg->x1;
 		eptinfo[nroute].starty = seg->y1;
 		eptinfo[nroute].startl = seg->layer;
-		eptinfo[nroute].starttype = seg->segtype;
 		eptinfo[nroute].res = 0.0;
 		eptinfo[nroute].cap = 0.0;
+
+		/* If a via, check if direction is up or down */
+		if (seg->segtype & ST_VIA) {
+		    if (seg->next && (seg->next->layer <= seg->layer))
+			eptinfo[nroute].startl++;
+		}
 	    }
 
 	    /* Segment end */
-	    for (seg = rt->segments; seg && seg->next; seg = seg->next);
+	    testl = eptinfo[nroute].startl;
+	    for (seg = rt->segments; seg && seg->next; seg = seg->next)
+		testl = seg->layer;
+
 	    if (seg != NULL) {
 		eptinfo[nroute].endx = seg->x2;
 		eptinfo[nroute].endy = seg->y2;
 		eptinfo[nroute].endl = seg->layer;
-		eptinfo[nroute].endtype = seg->segtype;
+
+		/* If a via, check if direction is up or down */
+		if (seg->segtype & ST_VIA) {
+		    if (testl <= seg->layer)
+			eptinfo[nroute].endl++;
+		}
 	    }
 	    nroute++;
 	}
@@ -592,19 +585,23 @@ int write_delays(char *filename)
 	j = 0;
 	for (rt = droutes; rt; rt = rt->next) {
 	    ROUTE testroute;
-	    int startx, starty, startl, starttype;
-	    int endx, endy, endl, endtype;
-	    int brkx, brky, brki, startcompat, endcompat;
+	    int startx, starty, startl;
+	    int endx, endy, endl;
+	    int brkx, brky, brkl, brki, startcompat, endcompat;
 	    int initial, final;
 	    int x1, y1, x2, y2;
 
 	    /* Check all segments (but not the endpoints) */
-	    for (seg = rt->segments; seg; seg = seg->next) {
+	    lastseg = NULL;
+	    for (seg = rt->segments; seg; lastseg = seg, seg = seg->next) {
 		initial = (seg == rt->segments) ? 1 : 0;
 		final = (seg->next == NULL) ? 1 : 0;
 
 		if (initial && (seg->segtype & ST_VIA)) continue;
-		if (final && (seg->segtype & ST_VIA)) continue;
+		if (final && (seg->segtype & ST_VIA) &&
+				((lastseg != rt->segments) ||
+				(!(rt->segments->segtype & ST_VIA))))
+		    continue;
 
 		x1 = seg->x1;
 		x2 = seg->x2;
@@ -635,8 +632,10 @@ int write_delays(char *filename)
 			    x2++;
 			else if (x1 < x2)
 			    x2--;
-			else
-			    continue;	/* shouldn't happen */
+			/* x1 == x2 here implies that the route is made of */
+			/* exactly two vias.  To continue would be to miss */
+			/* the center point between the vias.  This unique */
+			/* condition is type == ST_VIA, final == TRUE.	   */
 		    }
 		    else {
 			if (y1 > y2)
@@ -652,7 +651,6 @@ int write_delays(char *filename)
 		brki = -1;
 		for (i = 0; i < numroutes; i++) {
 		    if (eptinfo[i].route == rt) continue;
-		    if (eptinfo[i].endl == -2) continue;
 
 		    testroute = eptinfo[i].orig;
 		    if ((!(testroute->flags & RT_START_NODE)) &&
@@ -670,42 +668,33 @@ int write_delays(char *filename)
 		    startx = eptinfo[i].startx;
 		    starty = eptinfo[i].starty;
 		    startl = eptinfo[i].startl;
-		    starttype = eptinfo[i].starttype;
 		    endx = eptinfo[i].endx;
 		    endy = eptinfo[i].endy;
 		    endl = eptinfo[i].endl;
-		    endtype = eptinfo[i].endtype;
 
 		    /* Check various combinations of wire and via layers */
-		    if (seg->segtype & ST_WIRE) {
-			if (starttype & ST_WIRE)
-			    startcompat = (startl == seg->layer);
-			else
-			    startcompat = (startl == seg->layer)
-					|| (startl + 1 == seg->layer);
 
-			if (endtype & ST_WIRE)
+		    if (seg->segtype & ST_WIRE) {
+			startcompat = (startl == seg->layer);
+			endcompat = (endl == seg->layer);
+		    }
+		    else if (final) {
+			/* Unique condition: route is two vias.  Look	*/
+			/* only at the point between the vias.		*/
+			if (lastseg->layer < seg->layer) {
+			    startcompat = (startl == seg->layer);
 			    endcompat = (endl == seg->layer);
-			else
-			    endcompat = (endl == seg->layer)
-					|| (endl + 1 == seg->layer);
+			}
+			else {
+			    startcompat = (startl == seg->layer + 1);
+			    endcompat = (endl == seg->layer + 1);
+			}
 		    }
 		    else {
-			if (starttype & ST_WIRE)
-			    startcompat = (startl == seg->layer)
+			startcompat = (startl == seg->layer)
 					|| (startl == seg->layer + 1);
-			else
-			    startcompat = (startl == seg->layer)
-					|| (startl == seg->layer + 1)
-					|| (startl + 1 == seg->layer);
-
-			if (endtype & ST_WIRE)
-			    endcompat = (endl == seg->layer)
+			endcompat = (endl == seg->layer)
 					|| (endl == seg->layer + 1);
-			else
-			    endcompat = (endl == seg->layer)
-					|| (endl == seg->layer + 1)
-					|| (endl + 1 == seg->layer);
 		    }
 
 		    if (x1 == x2) {
@@ -715,6 +704,7 @@ int write_delays(char *filename)
 					starty <= y1) {
 				    brkx = startx;
 				    brky = starty;
+				    brkl = startl;
 				    y2 = brky;
 				    brki = i;
 				}
@@ -724,6 +714,7 @@ int write_delays(char *filename)
 					starty <= y2) {
 				    brkx = startx;
 				    brky = starty;
+				    brkl = startl;
 				    y2 = brky;
 				    brki = i;
 				}
@@ -735,6 +726,7 @@ int write_delays(char *filename)
 					endy <= y1) {
 				    brkx = endx;
 				    brky = endy;
+				    brkl = endl;
 				    y2 = brky;
 				    brki = i;
 				}
@@ -744,6 +736,7 @@ int write_delays(char *filename)
 					endy <= y2) {
 				    brkx = endx;
 				    brky = endy;
+				    brkl = endl;
 				    y2 = brky;
 				    brki = i;
 				}
@@ -757,6 +750,7 @@ int write_delays(char *filename)
 					startx <= x1) {
 				    brkx = startx;
 				    brky = starty;
+				    brkl = startl;
 				    x2 = brkx;
 				    brki = i;
 				}
@@ -766,6 +760,7 @@ int write_delays(char *filename)
 					startx <= x2) {
 				    brkx = startx;
 				    brky = starty;
+				    brkl = startl;
 				    x2 = brkx;
 				    brki = i;
 				}
@@ -777,6 +772,7 @@ int write_delays(char *filename)
 					endx <= x1) {
 				    brkx = endx;
 				    brky = endy;
+				    brkl = endl;
 				    x2 = brkx;
 				    brki = i;
 				}
@@ -786,6 +782,7 @@ int write_delays(char *filename)
 					endx <= x2) {
 				    brkx = endx;
 				    brky = endy;
+				    brkl = endl;
 				    x2 = brkx;
 				    brki = i;
 				}
@@ -793,27 +790,64 @@ int write_delays(char *filename)
 			}
 		    }
 		}
-		if ((brki >= 0) && (eptinfo[brki].endl != -2)) {
+		if (brki >= 0) {
 		    /* Disable this endpoint so it is not checked again */
 		    eptinfo[brki].endl = -2;
 
 		    /* Break route at this point */
-		    /* Make a copy of the segment where the break occurs */
+		    /* If route type is a wire, then make a copy of the	*/
+		    /* segment where the break occurs.  If a via, then	*/
+		    /* determine which side of the break the via goes	*/
+		    /* to.						*/
+
 		    newroute = (ROUTE)malloc(sizeof(struct route_));
-		    newseg = (SEG)malloc(sizeof(struct seg_));
-		    newseg->segtype = seg->segtype;
-		    newseg->x1 = brkx;
-		    newseg->y1 = brky;
-		    newseg->x2 = seg->x2;
-		    newseg->y2 = seg->y2;
-		    newseg->layer = seg->layer;
 
-		    newseg->next = seg->next;
-		    seg->next = NULL;
-		    seg->x2 = brkx;
-		    seg->y2 = brky;
+		    if (seg->segtype & ST_WIRE) {
+			newseg = (SEG)malloc(sizeof(struct seg_));
+			newseg->segtype = seg->segtype;
+			newseg->x1 = brkx;
+			newseg->y1 = brky;
+			newseg->x2 = seg->x2;
+			newseg->y2 = seg->y2;
+			newseg->layer = seg->layer;
+
+			newseg->next = seg->next;
+			seg->next = NULL;
+			seg->x2 = brkx;
+			seg->y2 = brky;
+
+			newroute->segments = newseg;
+		    }
+		    else if (lastseg == NULL) {
+			/* Via is the route before the break */
+			newroute->segments = seg->next;
+			seg->next = NULL;
+		    }
+		    else if (lastseg->layer <= seg->layer) {
+			if (brkl > seg->layer) {
+			    /* Via goes at the end of the route before the break */
+			    newroute->segments = seg->next;
+			    seg->next = NULL;
+			}
+			else {
+			    /* Via goes at the start of the route after the break */
+			    newroute->segments = seg;
+			    lastseg->next = NULL;
+			}
+		    }
+		    else {	/* (lastseg->layer > seg->layer) */
+			if (brkl > seg->layer) {
+			    /* Via goes at the start of the route after the break */
+			    newroute->segments = seg;
+			    lastseg->next = NULL;
+			}
+			else {
+			    /* Via goes at the end of the route before the break */
+			    newroute->segments = seg->next;
+			    seg->next = NULL;
+			}
+		    }
 
-		    newroute->segments = newseg;
 		    newroute->netnum = rt->netnum;
 		    newroute->flags = (u_char)0;
 		    newroute->next = rt->next;
@@ -862,7 +896,9 @@ int write_delays(char *filename)
 	    eptinfo[nroute].startx = seg->x1;
 	    eptinfo[nroute].starty = seg->y1;
 	    eptinfo[nroute].startl = seg->layer;
-	    eptinfo[nroute].starttype = seg->segtype;
+	    /* Check for via in down direction rather than the default up */
+	    if ((seg->segtype & ST_VIA) && seg->next && seg->next->layer <= seg->layer)
+		eptinfo[nroute].startl++;
 	    nodeptr = (seg->layer < Pinlayers) ?
 			NODEIPTR(seg->x1, seg->y1, seg->layer) : NULL;
 	    eptinfo[nroute].startnode = nodeptr ? nodeptr->nodesav : NULL;
@@ -874,7 +910,7 @@ int write_delays(char *filename)
 	    eptinfo[nroute].cap = 0.0;
 
 	    /* Look up node */
-	    if (nodeptr) {
+	    if (nodeptr && nodeptr->nodesav) {
 		g = FindGateNode(&NodeTable, nodeptr->nodesav, &i);
 		if (g && (g->gatetype->direction[i] == PORT_CLASS_OUTPUT)) {
 		    drivernodeidx = i;
@@ -908,13 +944,20 @@ int write_delays(char *filename)
 	    eptinfo[nroute].endx = seg->x2;
 	    eptinfo[nroute].endy = seg->y2;
 	    eptinfo[nroute].endl = seg->layer;
-	    eptinfo[nroute].endtype = seg->segtype;
-	    nodeptr = (seg->layer < Pinlayers) ?
-			NODEIPTR(seg->x2, seg->y2, seg->layer) : NULL;
+	    /* Check for via in down direction rather than default up */
+	    if (seg->segtype & ST_VIA) {
+		if (lastseg && lastseg->layer <= seg->layer)
+		    eptinfo[nroute].endl++;
+		else if (!lastseg && eptinfo[nroute].endl <= seg->layer)
+		    eptinfo[nroute].endl++;
+	    }
+
+	    nodeptr = (eptinfo[nroute].endl < Pinlayers) ?
+			NODEIPTR(seg->x2, seg->y2, eptinfo[nroute].endl) : NULL;
 	    eptinfo[nroute].endnode = nodeptr ? nodeptr->nodesav : NULL;
 
 	    /* Look up node */
-	    if (nodeptr) {
+	    if (nodeptr && nodeptr->nodesav) {
 		g = FindGateNode(&NodeTable, nodeptr->nodesav, &i);
 		if (g && (g->gatetype->direction[i] == PORT_CLASS_OUTPUT)) {
 		    drivernodeidx = i;
@@ -959,11 +1002,11 @@ int write_delays(char *filename)
 			drivergate->gatetype->node[drivernodeidx]);
 	    */
 	    if (!strcmp(drivergate->gatetype->node[drivernodeidx], "pin"))
-		fprintf(delayFile, " PIN/%s %d ",
-			drivergate->gatename, net->numnodes - 1);
+		fprintf(delayFile, "%s 1 PIN/%s %d ",
+			net->netname, drivergate->gatename, net->numnodes - 1);
 	    else
-		fprintf(delayFile, " %s/%s %d ",
-			drivergate->gatename,
+		fprintf(delayFile, "%s 1 %s/%s %d ",
+			net->netname, drivergate->gatename,
 			drivergate->gatetype->node[drivernodeidx],
 			net->numnodes - 1);
 
@@ -1012,6 +1055,7 @@ int write_delays(char *filename)
 
     free(lefrcvalues);
 
+    FreeNodeTable(&NodeTable);
     Tcl_DeleteHashTable(&NodeTable);
 
     return 0;
