@@ -151,10 +151,10 @@ int set_powerbus_to_net(int netnum)
     PROUTE *Pr;
 
     rval = 0;
-    if ((netnum == VDD_NET) || (netnum == GND_NET)) {
+    if ((netnum == VDD_NET) || (netnum == GND_NET) || (netnum == ANTENNA_NET)) {
        for (lay = 0; lay < Num_layers; lay++)
-          for (x = 0; x < NumChannelsX[lay]; x++)
-	     for (y = 0; y < NumChannelsY[lay]; y++)
+          for (x = 0; x < NumChannelsX; x++)
+	     for (y = 0; y < NumChannelsY; y++)
 		if ((OBSVAL(x, y, lay) & NETNUM_MASK) == netnum) {
 		   Pr = &OBS2VAL(x, y, lay);
 		   // Skip locations that have been purposefully disabled
@@ -370,7 +370,8 @@ int set_node_to_net(NODE node, int newflags, POINT *pushlist,
     /* net is a power bus, just return.					*/
 
     if ((node->taps == NULL) && (node->extend == NULL)) {
-       if ((node->netnum == VDD_NET) || (node->netnum == GND_NET))
+       if ((node->netnum == VDD_NET) || (node->netnum == GND_NET) ||
+		(node->netnum == ANTENNA_NET))
 	   return result;
     }
 
@@ -722,11 +723,14 @@ int set_route_to_net_recursive(NET net, ROUTE rt, int newflags,
 	    }
 	}
     }
-    else {
+    else if (rt->start.route) {
 	result = set_route_to_net_recursive(net, rt->start.route, newflags,
 			pushlist, bbox, stage);
 	if (result < 0) return result;
     }
+    else
+	Fprintf(stderr, "Error:  Route start information not recorded, cannot walk.\n");
+	
     if (rt->flags & RT_END_NODE) {
 	for (route = net->routes; route; route = route->next) {
 	    if (!(route->flags & RT_START_NODE) && (route->start.route == rt)) {
@@ -741,11 +745,13 @@ int set_route_to_net_recursive(NET net, ROUTE rt, int newflags,
 	    }
 	}
     }
-    else {
+    else if (rt->end.route) {
 	result = set_route_to_net_recursive(net, rt->end.route, newflags,
 			pushlist, bbox, stage);
 	if (result < 0) return result;
     }
+    else
+	Fprintf(stderr, "Error:  Route end information not recorded, cannot walk.\n");
     return result;
 }
 
@@ -879,13 +885,13 @@ NETLIST find_colliding(NET net, int *ripnum)
 	    while (1) {
 	       orignet = OBSVAL(x, y, lay) & ROUTED_NET_MASK;
 
-	       if (orignet == DRC_BLOCKAGE) {
+	       if ((orignet & DRC_BLOCKAGE) == DRC_BLOCKAGE) {
 
 		  /* If original position was a DRC-related blockage,	*/
 		  /* find out which net or nets would be in conflict.	*/
 
 		  if (needblock[lay] & (ROUTEBLOCKX | VIABLOCKX)) {
-		     if (x < NumChannelsX[lay] - 1) {
+		     if (x < NumChannelsX - 1) {
 		        orignet = OBSVAL(x + 1, y, lay) & ROUTED_NET_MASK;
 		        if (!(orignet & NO_NET)) {
 			   orignet &= NETNUM_MASK;
@@ -903,7 +909,7 @@ NETLIST find_colliding(NET net, int *ripnum)
 		     }
 		  }
 		  if (needblock[lay] & (ROUTEBLOCKY | VIABLOCKY)) {
-		     if (y < NumChannelsY[lay] - 1) {
+		     if (y < NumChannelsY - 1) {
 		        orignet = OBSVAL(x, y + 1, lay) & ROUTED_NET_MASK;
 		        if (!(orignet & NO_NET)) {
 			   orignet &= NETNUM_MASK;
@@ -1012,11 +1018,11 @@ void analyze_route_overwrite(int x, int y, int lay, int netnum)
 
     /* Check on all sides to see if position is orphaned */
 
-    if ((x < NumChannelsX[0] - 1) && (OBSVAL(x + 1, y, lay) & NETNUM_MASK) == netnum)
+    if ((x < NumChannelsX - 1) && (OBSVAL(x + 1, y, lay) & NETNUM_MASK) == netnum)
 	is_valid = TRUE;
     else if ((x > 0) && (OBSVAL(x - 1, y, lay) & NETNUM_MASK) == netnum)
 	is_valid = TRUE;
-    else if ((y < NumChannelsY[0] - 1) && (OBSVAL(x, y + 1, lay) & NETNUM_MASK) == netnum)
+    else if ((y < NumChannelsY - 1) && (OBSVAL(x, y + 1, lay) & NETNUM_MASK) == netnum)
 	is_valid = TRUE;
     else if ((y > 0) && (OBSVAL(x, y - 1, lay) & NETNUM_MASK) == netnum)
 	is_valid = TRUE;
@@ -1120,6 +1126,44 @@ void remove_routes(ROUTE netroutes, u_char flagged)
    }
 }
 
+/*--------------------------------------------------------------*/
+/* Clear a DRC blockage.  Normally, just remove the		*/
+/* DRC_BLOCKAGE flag from Obs[].  However, more than one net	*/
+/* may set a block on the position, so the lower four bits	*/
+/* (OBSTRUCT_MASK) holds the reference count.			*/
+/*--------------------------------------------------------------*/
+
+void clear_drc_blockage(x, y, lay)
+{
+    int blockcount;
+
+    blockcount = OBSVAL(x, y, lay) & OBSTRUCT_MASK;
+    OBSVAL(x, y, lay) &= ~OBSTRUCT_MASK;
+    if (blockcount > 0)
+	OBSVAL(x, y, lay) |= (blockcount - 1);
+    else
+	OBSVAL(x, y, lay) &= ~DRC_BLOCKAGE;
+}
+
+/*--------------------------------------------------------------*/
+/*--------------------------------------------------------------*/
+
+void set_drc_blockage(x, y, lay)
+{
+    int blockcount, obsval;
+
+    obsval = OBSVAL(x, y, lay);
+    if ((obsval & DRC_BLOCKAGE) == DRC_BLOCKAGE) {
+	blockcount = OBSVAL(x, y, lay) & OBSTRUCT_MASK;
+	OBSVAL(x, y, lay) &= ~OBSTRUCT_MASK;
+	OBSVAL(x, y, lay) |= (blockcount + 1);
+    }
+    else if ((obsval & NETNUM_MASK) == 0) {
+	OBSVAL(x, y, lay) &= ~OBSTRUCT_MASK;
+	OBSVAL(x, y, lay) |= DRC_BLOCKAGE;
+    }
+}
+
 /*--------------------------------------------------------------*/
 /* ripup_net ---						*/
 /*								*/
@@ -1169,7 +1213,8 @@ u_char ripup_net(NET net, u_char restore, u_char flagged, u_char retain)
 		     // net and route associated with the incorrect net.
 		     analyze_route_overwrite(x, y, lay, oldnet);
 
-		     // return FALSE;	// Something went wrong
+		     if ((rt == NULL) || (rt->segments == NULL))
+			 return FALSE;	// Something went wrong
 	          }
 
 	          // Reset the net number to zero along this route for
@@ -1197,20 +1242,20 @@ u_char ripup_net(NET net, u_char restore, u_char flagged, u_char retain)
 		  if (needblock[lay] & (ROUTEBLOCKX | VIABLOCKX)) {
 		     if ((x > 0) && ((OBSVAL(x - 1, y, lay) &
 				DRC_BLOCKAGE) == DRC_BLOCKAGE))
-			OBSVAL(x - 1, y, lay) &= ~DRC_BLOCKAGE;
-		     else if ((x < NumChannelsX[lay] - 1) &&
+			clear_drc_blockage(x - 1, y, lay);
+		     else if ((x < NumChannelsX - 1) &&
 				((OBSVAL(x + 1, y, lay) &
 				DRC_BLOCKAGE) == DRC_BLOCKAGE))
-			OBSVAL(x + 1, y, lay) &= ~DRC_BLOCKAGE;
+			clear_drc_blockage(x + 1, y, lay);
 		  }
 		  if (needblock[lay] & (ROUTEBLOCKY | VIABLOCKY)) {
 		     if ((y > 0) && ((OBSVAL(x, y - 1, lay) &
 				DRC_BLOCKAGE) == DRC_BLOCKAGE))
-			OBSVAL(x, y - 1, lay) &= ~DRC_BLOCKAGE;
-		     else if ((y < NumChannelsY[lay] - 1) &&
+			clear_drc_blockage(x, y - 1, lay);
+		     else if ((y < NumChannelsY - 1) &&
 				((OBSVAL(x, y + 1, lay) &
 				DRC_BLOCKAGE) == DRC_BLOCKAGE))
-			OBSVAL(x, y + 1, lay) &= ~DRC_BLOCKAGE;
+			clear_drc_blockage(x, y + 1, lay);
 		  }
 	       }
 
@@ -1379,7 +1424,7 @@ POINT eval_pt(GRIDP *ept, u_char flags, u_char stage)
 	  Pr->prdata.cost = MAXRT;
 	  thiscost += ConflictCost;
        }
-       else if (stage && (netnum == DRC_BLOCKAGE)) {
+       else if (stage && ((netnum & DRC_BLOCKAGE) == DRC_BLOCKAGE)) {
 	  if ((newpt.lay < Pinlayers) && nodeptr && (nodeptr->nodesav != NULL))
 	     return NULL;		// But cannot route over terminals!
 
@@ -1390,7 +1435,7 @@ POINT eval_pt(GRIDP *ept, u_char flags, u_char stage)
 	  // route here if any of them are on the noripup list.
 
 	  if (needblock[newpt.lay] & (ROUTEBLOCKX | VIABLOCKX)) {
-	     if (newpt.x < NumChannelsX[newpt.lay] - 1) {
+	     if (newpt.x < NumChannelsX - 1) {
 	        netnum = OBSVAL(newpt.x + 1, newpt.y, newpt.lay) & ROUTED_NET_MASK;
 	        if (!(netnum & NO_NET)) {
 		   netnum &= NETNUM_MASK;
@@ -1415,7 +1460,7 @@ POINT eval_pt(GRIDP *ept, u_char flags, u_char stage)
 	     }
 	  } 
 	  if (needblock[newpt.lay] & (ROUTEBLOCKY | VIABLOCKY)) {
-	     if (newpt.y < NumChannelsY[newpt.lay] - 1) {
+	     if (newpt.y < NumChannelsY - 1) {
 	        netnum = OBSVAL(newpt.x, newpt.y + 1, newpt.lay) & ROUTED_NET_MASK;
 	        if (!(netnum & NO_NET)) {
 		   netnum &= NETNUM_MASK;
@@ -1578,20 +1623,16 @@ void writeback_segment(SEG seg, int netnum)
       dir = OBSVAL(seg->x1, seg->y1, seg->layer + 1) & (BLOCKED_MASK | PINOBSTRUCTMASK);
       OBSVAL(seg->x1, seg->y1, seg->layer + 1) = netnum | dir;
       if (needblock[seg->layer + 1] & VIABLOCKX) {
-	 if ((seg->x1 < (NumChannelsX[seg->layer + 1] - 1)) &&
-		(OBSVAL(seg->x1 + 1, seg->y1, seg->layer + 1) & NETNUM_MASK) == 0)
- 	    OBSVAL(seg->x1 + 1, seg->y1, seg->layer + 1) = DRC_BLOCKAGE;
-	 if ((seg->x1 > 0) &&
-		(OBSVAL(seg->x1 - 1, seg->y1, seg->layer + 1) & NETNUM_MASK) == 0)
-	    OBSVAL(seg->x1 - 1, seg->y1, seg->layer + 1) = DRC_BLOCKAGE;
+	 if (seg->x1 < (NumChannelsX - 1))
+ 	    set_drc_blockage(seg->x1 + 1, seg->y1, seg->layer + 1);
+	 if (seg->x1 > 0)
+	    set_drc_blockage(seg->x1 - 1, seg->y1, seg->layer + 1);
       }
       if (needblock[seg->layer + 1] & VIABLOCKY) {
-	 if ((seg->y1 < (NumChannelsY[seg->layer + 1] - 1)) &&
-		(OBSVAL(seg->x1, seg->y1 + 1, seg->layer + 1) & NETNUM_MASK) == 0)
-	    OBSVAL(seg->x1, seg->y1 + 1, seg->layer + 1) = DRC_BLOCKAGE;
-	 if ((seg->y1 > 0) &&
-		(OBSVAL(seg->x1, seg->y1 - 1, seg->layer + 1) & NETNUM_MASK) == 0)
-	    OBSVAL(seg->x1, seg->y1 - 1, seg->layer + 1) = DRC_BLOCKAGE;
+	 if (seg->y1 < (NumChannelsY - 1))
+	    set_drc_blockage(seg->x1, seg->y1 + 1, seg->layer + 1);
+	 if (seg->y1 > 0)
+	    set_drc_blockage(seg->x1, seg->y1 - 1, seg->layer + 1);
       }
 
       // If position itself is an offset route, then make the route position
@@ -1605,23 +1646,23 @@ void writeback_segment(SEG seg, int netnum)
 	 lnode = NODEIPTR(seg->x1, seg->y1, seg->layer);
 	 dist = lnode->offset;
 	 if (lnode->flags & NI_OFFSET_EW) {
-	    if ((dist > 0) && (seg->x1 < (NumChannelsX[seg->layer] - 1))) {
-	       OBSVAL(seg->x1 + 1, seg->y1, seg->layer) |= DRC_BLOCKAGE;
-	       OBSVAL(seg->x1 + 1, seg->y1, seg->layer + 1) |= DRC_BLOCKAGE;
+	    if ((dist > 0) && (seg->x1 < (NumChannelsX - 1))) {
+	       set_drc_blockage(seg->x1 + 1, seg->y1, seg->layer);
+	       set_drc_blockage(seg->x1 + 1, seg->y1, seg->layer + 1);
 	    }
 	    if ((dist < 0) && (seg->x1 > 0)) {
-	       OBSVAL(seg->x1 - 1, seg->y1, seg->layer) |= DRC_BLOCKAGE;
-	       OBSVAL(seg->x1 - 1, seg->y1, seg->layer + 1) |= DRC_BLOCKAGE;
+	       set_drc_blockage(seg->x1 - 1, seg->y1, seg->layer);
+	       set_drc_blockage(seg->x1 - 1, seg->y1, seg->layer + 1);
 	    }
 	 }
 	 else if (lnode->flags & NI_OFFSET_NS) {
-	    if ((dist > 0) && (seg->y1 < (NumChannelsY[seg->layer] - 1))) {
-	       OBSVAL(seg->x1, seg->y1 + 1, seg->layer) |= DRC_BLOCKAGE;
-	       OBSVAL(seg->x1, seg->y1 + 1, seg->layer + 1) |= DRC_BLOCKAGE;
+	    if ((dist > 0) && (seg->y1 < (NumChannelsY - 1))) {
+	       set_drc_blockage(seg->x1, seg->y1 + 1, seg->layer);
+	       set_drc_blockage(seg->x1, seg->y1 + 1, seg->layer + 1);
 	    }
 	    if ((dist < 0) && (seg->y1 > 0)) {
-	       OBSVAL(seg->x1, seg->y1 - 1, seg->layer) |= DRC_BLOCKAGE;
-	       OBSVAL(seg->x1, seg->y1 - 1, seg->layer + 1) |= DRC_BLOCKAGE;
+	       set_drc_blockage(seg->x1, seg->y1 - 1, seg->layer);
+	       set_drc_blockage(seg->x1, seg->y1 - 1, seg->layer + 1);
 	    }
 	 }
       }
@@ -1631,12 +1672,10 @@ void writeback_segment(SEG seg, int netnum)
       dir = OBSVAL(i, seg->y1, seg->layer) & (BLOCKED_MASK | PINOBSTRUCTMASK);
       OBSVAL(i, seg->y1, seg->layer) = netnum | dir;
       if (needblock[seg->layer] & ROUTEBLOCKY) {
-         if ((seg->y1 < (NumChannelsY[seg->layer] - 1)) &&
-		(OBSVAL(i, seg->y1 + 1, seg->layer) & NETNUM_MASK) == 0)
-	    OBSVAL(i, seg->y1 + 1, seg->layer) = DRC_BLOCKAGE;
-	 if ((seg->y1 > 0) &&
-		(OBSVAL(i, seg->y1 - 1, seg->layer) & NETNUM_MASK) == 0)
-	    OBSVAL(i, seg->y1 - 1, seg->layer) = DRC_BLOCKAGE;
+         if (seg->y1 < (NumChannelsY - 1))
+	    set_drc_blockage(i, seg->y1 + 1, seg->layer);
+	 if (seg->y1 > 0)
+	    set_drc_blockage(i, seg->y1 - 1, seg->layer);
       }
 
       // Check position on each side for an offset tap on a different net, and
@@ -1647,14 +1686,14 @@ void writeback_segment(SEG seg, int netnum)
       // cause a DRC error.  Could be refined. . .
 
       layer = (seg->layer == 0) ? 0 : seg->layer - 1;
-      if (seg->y1 < (NumChannelsY[layer] - 1)) {
+      if (seg->y1 < (NumChannelsY - 1)) {
 	 sobs = OBSVAL(i, seg->y1 + 1, layer);
 	 if ((sobs & OFFSET_TAP) && !(sobs & ROUTED_NET)) {
 	    lnode = NODEIPTR(i, seg->y1 + 1, layer);
 	    if (lnode->flags & NI_OFFSET_NS) {
 	       dist = lnode->offset;
 	       if (dist < 0) {
-		  OBSVAL(i, seg->y1 + 1, layer) |= DRC_BLOCKAGE;
+		  set_drc_blockage(i, seg->y1 + 1, layer);
 	       }
 	    }
 	 }
@@ -1666,37 +1705,50 @@ void writeback_segment(SEG seg, int netnum)
 	    if (lnode->flags & NI_OFFSET_NS) {
 	       dist = lnode->offset;
 	       if (dist > 0) {
-		  OBSVAL(i, seg->y1 - 1, layer) |= DRC_BLOCKAGE;
+		  set_drc_blockage(i, seg->y1 - 1, layer);
 	       }
 	    }
 	 }
       }
       if (i == seg->x2) break;
    }
+
+   /* Check top of route for vertical routes */
+
+   if (seg->y1 != seg->y2) {
+      dir = OBSVAL(seg->x2, seg->y2, seg->layer) & (BLOCKED_MASK | PINOBSTRUCTMASK);
+      OBSVAL(seg->x2, seg->y2, seg->layer) = netnum | dir;
+      if (needblock[seg->layer] & ROUTEBLOCKY) {
+         if (seg->y2 < (NumChannelsY - 1))
+	    set_drc_blockage(seg->x2, seg->y2 + 1, seg->layer);
+	 if (seg->y2 > 0)
+	    set_drc_blockage(seg->x2, seg->y2 - 1, seg->layer);
+      }
+   }
+
+
    for (i = seg->y1; ; i += (seg->y2 > seg->y1) ? 1 : -1) {
       dir = OBSVAL(seg->x1, i, seg->layer) & (BLOCKED_MASK | PINOBSTRUCTMASK);
       OBSVAL(seg->x1, i, seg->layer) = netnum | dir;
       if (needblock[seg->layer] & ROUTEBLOCKX) {
-	 if ((seg->x1 < (NumChannelsX[seg->layer] - 1)) &&
-		(OBSVAL(seg->x1 + 1, i, seg->layer) & NETNUM_MASK) == 0)
-	    OBSVAL(seg->x1 + 1, i, seg->layer) = DRC_BLOCKAGE;
-	 if ((seg->x1 > 0) &&
-		(OBSVAL(seg->x1 - 1, i, seg->layer) & NETNUM_MASK) == 0)
-	    OBSVAL(seg->x1 - 1, i, seg->layer) = DRC_BLOCKAGE;
+	 if (seg->x1 < (NumChannelsX - 1))
+	    set_drc_blockage(seg->x1 + 1, i, seg->layer);
+	 if (seg->x1 > 0)
+	    set_drc_blockage(seg->x1 - 1, i, seg->layer);
       }
 
       // Check position on each side for an offset tap on a different net, and
       // mark the position unroutable (see above).
 
       layer = (seg->layer == 0) ? 0 : seg->layer - 1;
-      if (seg->x1 < (NumChannelsX[layer] - 1)) {
+      if (seg->x1 < (NumChannelsX - 1)) {
 	 sobs = OBSVAL(seg->x1 + 1, i, layer);
 	 if ((sobs & OFFSET_TAP) && !(sobs & ROUTED_NET)) {
 	    lnode = NODEIPTR(seg->x1 + 1, i, layer);
 	    if (lnode->flags & NI_OFFSET_EW) {
 	       dist = lnode->offset;
 	       if (dist < 0) {
-		  OBSVAL(seg->x1 + 1, i, layer) |= DRC_BLOCKAGE;
+		  set_drc_blockage(seg->x1 + 1, i, layer);
 	       }
 	    }
 	 }
@@ -1708,13 +1760,26 @@ void writeback_segment(SEG seg, int netnum)
 	    if (lnode->flags & NI_OFFSET_EW) {
 	       dist = lnode->offset;
 	       if (dist > 0) {
-		  OBSVAL(seg->x1 - 1, i, layer) |= DRC_BLOCKAGE;
+		  set_drc_blockage(seg->x1 - 1, i, layer);
 	       }
 	    }
 	 }
       }
       if (i == seg->y2) break;
    }
+
+   /* Check end of route for horizontal routes */
+
+   if (seg->x1 != seg->x2) {
+      dir = OBSVAL(seg->x2, seg->y2, seg->layer) & (BLOCKED_MASK | PINOBSTRUCTMASK);
+      OBSVAL(seg->x2, seg->y2, seg->layer) = netnum | dir;
+      if (needblock[seg->layer] & ROUTEBLOCKX) {
+	 if (seg->x2 < (NumChannelsX - 1))
+	    set_drc_blockage(seg->x2 + 1, seg->y2, seg->layer);
+	 if (seg->x2 > 0)
+	    set_drc_blockage(seg->x2 - 1, seg->y2, seg->layer);
+      }
+   }
 }
 
 /*--------------------------------------------------------------*/
@@ -1735,19 +1800,56 @@ route_set_connections(net, route)
    u_char   found, match;
    int	    x, y;
 
+   /* Reset start/end node flags */
+   route->flags &= ~(RT_START_NODE | RT_END_NODE);
+
    /* Does first route segment connect to a node? */
 
    seg = route->segments;
    found = FALSE;
    if (seg->layer < Pinlayers) {
       lnode = NODEIPTR(seg->x1, seg->y1, seg->layer);
-      if (lnode != NULL) {
+      if ((lnode != NULL) && (lnode->nodesav != NULL)) {
 	 route->start.node = lnode->nodesav;
 	 route->flags |= RT_START_NODE;
 	 found = TRUE;
       }
    }
 
+   /* For routes that were just imported from DEF, check if the	*/
+   /* route overshot a pin terminal by rounding off a non-grid	*/
+   /* coordinate.						*/
+
+   if (!found && (route->flags & RT_CHECK) && (seg->layer < Pinlayers)) {
+      if (seg->x1 == seg->x2) {
+	 x = seg->x1;
+	 if (seg->y1 < seg->y2)
+	    y = seg->y1 + 1;
+         else
+	    y = seg->y1 - 1;
+      }
+      else {
+	 y = seg->y1;
+	 if (seg->x1 < seg->x2)
+	    x = seg->x1 + 1;
+         else
+	    x = seg->x1 - 1;
+      }
+      lnode = NODEIPTR(x, y, seg->layer);
+      if ((lnode != NULL) && (lnode->nodesav != NULL) &&
+		(lnode->nodesav->netnum == net->netnum) &&
+		((x != seg->x2) || (y != seg->y2))) {
+	 route->start.node = lnode->nodesav;
+	 route->flags |= RT_START_NODE;
+	 found = TRUE;
+	 /* Diagnostic */
+	 Fprintf(stderr, "Coordinate %d %d corrected to %d %d\n",
+		seg->x1, seg->y1, x, y);
+	 seg->x1 = x;
+	 seg->y1 = y;
+      }
+   }
+
    /* Does first route segment connect to a route? */
 
    if (!found) {
@@ -1785,7 +1887,8 @@ route_set_connections(net, route)
    }
 
    if (!found) {
-      Fprintf(stderr, "Error:  Failure to find route start node/route!\n");
+      Fprintf(stderr, "Error:  Failure to find route start node/route on net %s!\n",
+		net->netname);
    }
 
    /* Does last route segment connect to a node? */
@@ -1799,12 +1902,46 @@ route_set_connections(net, route)
       for (; seg->next; seg = seg->next);
       if (seg->layer < Pinlayers) {
          lnode = NODEIPTR(seg->x2, seg->y2, seg->layer);
-         if (lnode != NULL) {
+         if ((lnode != NULL) && (lnode->nodesav != NULL)) {
 	    route->end.node = lnode->nodesav;
 	    route->flags |= RT_END_NODE;
 	    found = TRUE;
 	 }
       }
+
+      /* For routes that were just imported from DEF, check if	*/
+      /* the route overshot a pin terminal by rounding off a	*/
+      /* non-grid coordinate.					*/
+
+      if (!found && (route->flags & RT_CHECK) && (seg->layer < Pinlayers)) {
+         if (seg->x1 == seg->x2) {
+	    x = seg->x2;
+	    if (seg->y1 < seg->y2)
+	       y = seg->y2 - 1;
+            else
+	       y = seg->y2 + 1;
+         }
+         else {
+	    y = seg->y2;
+	    if (seg->x1 < seg->x2)
+	       x = seg->x2 - 1;
+            else
+	       x = seg->x2 + 1;
+         }
+         lnode = NODEIPTR(x, y, seg->layer);
+         if ((lnode != NULL) && (lnode->nodesav != NULL) &&
+			(lnode->nodesav->netnum == net->netnum) &&
+			((x != seg->x1) || (y != seg->y1))) {
+	    route->start.node = lnode->nodesav;
+	    route->flags |= RT_END_NODE;
+	    found = TRUE;
+	    /* Diagnostic */
+	    Fprintf(stderr, "Coordinate %d %d corrected to %d %d\n",
+			seg->x2, seg->y2, x, y);
+	    seg->x2 = x;
+	    seg->y2 = y;
+         }
+      }
    }
 
    /* Does last route segment connect to a route? */
@@ -1844,8 +1981,12 @@ route_set_connections(net, route)
    }
 
    if (!found) {
-      Fprintf(stderr, "Error:  Failure to find route end node/route!\n");
+      Fprintf(stderr, "Error:  Failure to find route end node/route on net %s!\n",
+		net->netname);
    }
+
+   /* Clear RT_CHECK flag after processing */
+   if (route->flags & RT_CHECK) route->flags &= ~RT_CHECK;
 }
 
 /*--------------------------------------------------------------*/
@@ -1957,6 +2098,8 @@ int commit_proute(ROUTE rt, GRIDP *ept, u_char stage)
 	    lrppre = lrprev->next;
 	    if (lrppre == NULL) break;
 	    stackheight = 0;
+	    /* Advance lrcur if jogs were inserted */
+	    while (lrprev != lrcur->next) lrcur = lrcur->next;
 	    a = lrcur;
 	    b = lrprev;
 	    while (a->layer != b->layer) {
@@ -1980,7 +2123,7 @@ int commit_proute(ROUTE rt, GRIDP *ept, u_char stage)
 	       // lowest cost, and make sure the position below that
 	       // is available.
 
-	       if (cx < NumChannelsX[cl] - 1) {
+	       if (cx < NumChannelsX - 1) {
 	          dx = cx + 1;	// Check to the right
 	          pri = &OBS2VAL(dx, cy, cl);
 	          pflags = pri->flags;
@@ -2049,7 +2192,7 @@ int commit_proute(ROUTE rt, GRIDP *ept, u_char stage)
 		  }
 	       }
 
-	       if (cy < NumChannelsY[cl] - 1) {
+	       if (cy < NumChannelsY - 1) {
 	          dy = cy + 1;	// Check north
 	          pri = &OBS2VAL(cx, dy, cl);
 	          pflags = pri->flags;
@@ -2171,7 +2314,7 @@ int commit_proute(ROUTE rt, GRIDP *ept, u_char stage)
 	          mincost = MAXRT;
 	          dl = lrprev->layer;
 
-		  if (cx < NumChannelsX[cl] - 1) {
+		  if (cx < NumChannelsX - 1) {
 	             dx = cx + 1;	// Check to the right
 	             pri = &OBS2VAL(dx, cy, cl);
 	             pflags = pri->flags;
@@ -2217,7 +2360,7 @@ int commit_proute(ROUTE rt, GRIDP *ept, u_char stage)
 		     }
 	          }
 
-		  if (cy < NumChannelsY[cl] - 1) {
+		  if (cy < NumChannelsY - 1) {
 	             dy = cy + 1;	// Check north
 	             pri = &OBS2VAL(cx, dy, cl);
 	             pflags = pri->flags;
@@ -2349,6 +2492,224 @@ int commit_proute(ROUTE rt, GRIDP *ept, u_char stage)
 	 }
       }
    }
+
+   /* Handle minimum metal area rule for stacked contacts */
+
+   else if (StackedContacts > 0) {
+
+      /* Search for all stacked contacts.  For each via position inside	*/
+      /* a stack, excluding top and bottom, determine if a minimum 	*/
+      /* metal area is violated.  If so, then find an unused		*/
+      /* neighboring grid and extend to that side.  If no unused	*/
+      /* neighboring grids exist, then fail.				*/
+
+      POINT lrppre;
+      PROUTE *pri, *pri2;
+      int violations = 1, checks;
+      int i, cx, cy, cl, o, orient;
+      int mincost, minx = -1, miny = -1, collide, cost;
+      double min_area[MAX_LAYERS];
+      u_char need_check[MAX_LAYERS];
+
+      /* Register the minimum metal area rule for each via base layer */
+
+      checks = 0;
+      for (i = 0; i < Num_layers; i++) {
+	 min_area[i] = LefGetRouteMinArea(i);
+	 if (i == Num_layers - 1) {
+	    if (min_area[i] <= LefGetViaWidth(i - 1, i, 0) * LefGetViaWidth(i - 1, i, 1))
+	       need_check[i] = (u_char)0;
+	    else {
+	       need_check[i] = (u_char)1;
+	       checks++;
+	    }
+	 }
+	 else {
+	    if (min_area[i] <= LefGetViaWidth(i, i, 0) * LefGetViaWidth(i, i, 1))
+	       need_check[i] = (u_char)0;
+	    else {
+	       need_check[i] = (u_char)1;
+	       checks++;
+	    }
+	 }
+      }
+
+      if (checks == 0) violations = 0;	// No minimum area violations possible.
+
+      while (violations != 0) {	// Keep doing until all min. area violations are gone
+	 violations = 0;
+	 lrcur = lrend;
+	 lrprev = lrend->next;
+
+	 while (lrprev != NULL) {
+	    lrppre = lrprev->next;
+	    if (lrppre == NULL) break;
+	    collide = FALSE;
+	    while ((lrcur->layer != lrprev->layer) && (lrprev->layer != lrppre->layer)
+			&& (need_check[lrprev->layer] == (u_char)1)) {
+	       // Isolated via inside stack found
+	
+	       violations++;
+	       cx = lrprev->x1;
+	       cy = lrprev->y1;
+	       cl = lrprev->layer;
+	       mincost = MAXRT;
+
+	       // Check all four positions around the contact for a
+	       // free position to occupy.  Positions in the preferred
+	       // direction of the route layer take precedence.
+
+	       o = LefGetRouteOrientation(cl);
+	       for (orient = 0; orient < 2; orient++) {
+		  if (o == orient) {
+		     if (cy < NumChannelsY - 1) {
+			dy = cy + 1;	// Check north
+			pri = &OBS2VAL(cx, dy, cl);
+			pflags = pri->flags;
+			cost = pri->prdata.cost;
+			if (collide && !(pflags & (PR_COST | PR_SOURCE)) &&
+					(pri->prdata.net < MAXNETNUM)) {
+			   pflags = 0;
+			   cost = ConflictCost;
+			}
+			if (pflags & PR_COST) {
+			   pflags &= ~PR_COST;
+			   if ((pflags & PR_PRED_DMASK) != PR_PRED_NONE &&
+					cost < mincost) {
+			      mincost = cost;
+			      minx = cx;
+			      miny = dy;
+			   }
+			}
+		     }
+
+		     if (cy > 0) {
+			dy = cy - 1;	// Check south
+			pri = &OBS2VAL(cx, dy, cl);
+			pflags = pri->flags;
+			cost = pri->prdata.cost;
+			if (collide && !(pflags & (PR_COST | PR_SOURCE)) &&
+					(pri->prdata.net < MAXNETNUM)) {
+			   pflags = 0;
+			   cost = ConflictCost;
+			}
+			if (pflags & PR_COST) {
+			   pflags &= ~PR_COST;
+			   if ((pflags & PR_PRED_DMASK) != PR_PRED_NONE &&
+					cost < mincost) {
+			      mincost = cost;
+			      minx = cx;
+			      miny = dy;
+			   }
+			}
+		     }
+		  }
+		  else {
+		     if (cx < NumChannelsX - 1) {
+			dx = cx + 1;	// Check to the right
+			pri = &OBS2VAL(dx, cy, cl);
+			pflags = pri->flags;
+			cost = pri->prdata.cost;
+			if (collide && !(pflags & (PR_COST | PR_SOURCE)) &&
+					(pri->prdata.net < MAXNETNUM)) {
+			   pflags = 0;
+			   cost = ConflictCost;
+			}
+			if (pflags & PR_COST) {
+			   pflags &= ~PR_COST;
+			   if ((pflags & PR_PRED_DMASK) != PR_PRED_NONE &&
+					cost < mincost) {
+			      mincost = cost;
+			      minx = dx;
+			      miny = cy;
+			   }
+			}
+		     }
+
+		     if (cx > 0) {
+			dx = cx - 1;	// Check to the left
+			pri = &OBS2VAL(dx, cy, cl);
+			pflags = pri->flags;
+			cost = pri->prdata.cost;
+			if (collide && !(pflags & (PR_COST | PR_SOURCE)) &&
+				(pri->prdata.net < MAXNETNUM)) {
+			   pflags = 0;
+			   cost = ConflictCost;
+			}
+			if (pflags & PR_COST) {
+			   pflags &= ~PR_COST;
+			   if ((pflags & PR_PRED_DMASK) != PR_PRED_NONE &&
+					cost < mincost) {
+			      mincost = cost;
+			      minx = dx;
+			      miny = cy;
+			   }
+			}
+		     }
+		  }
+	       }
+
+	       // Was there an available position?  If so, modify
+	       // records to route to and from this position.
+
+	       if (mincost < MAXRT) {
+
+		  newlr = (POINT)malloc(sizeof(struct point_));
+		  newlr->x1 = minx;
+		  newlr->y1 = miny;
+		  newlr->layer = cl;
+
+		  newlr2 = (POINT)malloc(sizeof(struct point_));
+		  newlr2->x1 = cx;
+		  newlr2->y1 = cy;
+		  newlr2->layer = cl;
+
+		  lrprev->next = newlr;
+		  newlr->next = newlr2;
+		  newlr2->next = lrppre;
+
+		  break;	/* Found a solution;  done. */
+	       }
+
+	       else if (stage == 0) {
+		  // On the first stage, we call it an error and move
+		  // on to the next net.  This is a bit conservative,
+		  // but it happens rarely.
+
+		  if (Verbose > 0)
+		     Fprintf(stderr, "Failed to reserve sufficient area "
+				"at grid point %d %d.\n",
+				lrcur->x1, lrcur->y1);
+		  violations = 0;
+		  rval = 0;
+		  goto cleanup;
+	       }
+	       else {
+		  if (collide == TRUE) {
+		     Fprintf(stderr, "Failed to reserve sufficient area "
+				"at grid point %d %d;  position may "
+				"not be routable.\n",
+				lrcur->x1, lrcur->y1);
+		     violations = 0;
+		     rval = 0;
+		     goto cleanup;
+		  }
+
+		  // On the second stage, we will run through the
+		  // search again, but allow overwriting other
+		  // nets, which will be treated like other colliding
+		  // nets in the regular route path search.
+
+		  collide = TRUE;
+	       }
+	    }
+	    lrcur = lrprev;
+	    lrprev = lrppre;
+	 }
+      }
+   }
+
+   /* Done handling stacked contacts */
  
    lrend = lrtop;
    lrcur = lrtop;
@@ -2575,7 +2936,10 @@ int writeback_route(ROUTE rt)
       lay2 = (seg->segtype & ST_VIA) ? seg->layer + 1 : seg->layer;
 
       dir1 = OBSVAL(seg->x1, seg->y1, seg->layer) & PINOBSTRUCTMASK;
-      dir2 = OBSVAL(seg->x2, seg->y2, lay2) & PINOBSTRUCTMASK;
+      if (lay2 < Num_layers)
+	 dir2 = OBSVAL(seg->x2, seg->y2, lay2) & PINOBSTRUCTMASK;
+      else
+	 dir2 = 0;
 
       writeback_segment(seg, netnum);
 
