@@ -8,14 +8,13 @@
 
 #define GND_NET		 1
 #define VDD_NET		 2
-#define MIN_NET_NUMBER   3
+#define ANTENNA_NET	 3
+#define MIN_NET_NUMBER   4
 
-void create_netorder(u_char method);
 void find_bounding_box(NET net);
 void defineRouteTree(NET);
-void print_nodes(char *filename);
-void print_nlnets(char *filename);
-void count_reachable_taps();
+DPOINT is_testpoint(int, int, GATE, int, DSEG);
+void count_reachable_taps(u_char);
 void check_variable_pitch(int, int *, int *);
 void create_obstructions_from_variable_pitch(void);
 void count_pinlayers(void);
