@@ -52,6 +52,7 @@ u_char needblock[MAX_LAYERS];
 
 char *vddnet = NULL;
 char *gndnet = NULL;
+char *antenna_cell = NULL;
 
 int    Numnets = 0;
 int    Pinlayers = 0;
@@ -62,12 +63,44 @@ u_char maskMode = MASK_AUTO;
 u_char mapType = MAP_OBSTRUCT | DRAW_ROUTES;
 u_char ripLimit = 10;	// Fail net rather than rip up more than
 			// this number of other nets.
+u_char unblockAll = FALSE;
 
 char *DEFfilename = NULL;
 char *delayfilename = NULL;
 
+DPOINT testpoint = NULL;	// used for debugging route problems
+
 ScaleRec Scales;	// record of input and output scales
 
+/*--------------------------------------------------------------*/
+/* Upate the output scale factor.  It has to be a valid DEF	*/
+/* scale factor and it has to be a multiple of the given scale	*/
+/* factor.							*/
+/*--------------------------------------------------------------*/
+
+void
+update_mscale(int mscale)
+{
+    static int valid_mscales[] = {100, 200, 1000, 2000, 10000, 20000};
+    int nscales = sizeof(valid_mscales) / sizeof(valid_mscales[0]);
+    int mscale2, i;
+
+    if (mscale == 0) return;
+
+    if ((Scales.mscale % mscale) != 0) {
+	// Check valid scale values;  if none is appropriate, don't update
+	for (i = 0; i < nscales; i++) {
+	    mscale2 = valid_mscales[i];
+	    if (mscale2 <= Scales.mscale) continue;
+	    if ((mscale2 % mscale) == 0) {
+		Scales.mscale = mscale2;
+		break;
+	    }
+	}
+    }
+}
+
+
 /*--------------------------------------------------------------*/
 /* Check track pitch and set the number of channels (may be	*/
 /* called from DefRead)						*/
@@ -80,33 +113,33 @@ int set_num_channels(void)
     NODE node;
     DPOINT ctap, ltap, ntap;
 
-    if (NumChannelsX[0] != 0) return 0;	/* Already been called */
+    if (NumChannelsX != 0) return 0;	/* Already been called */
 
-    for (i = 0; i < Num_layers; i++) {
-	if (PitchX[i] == 0.0 || PitchY[i] == 0.0) {
-	    Fprintf(stderr, "Have a 0 pitch for layer %d (of %d).  "
-			"Exit.\n", i + 1, Num_layers);
-	    return (-3);
-	}
-	NumChannelsX[i] = (int)(1.5 + (Xupperbound - Xlowerbound) / PitchX[i]);
-	NumChannelsY[i] = (int)(1.5 + (Yupperbound - Ylowerbound) / PitchY[i]);
-	if ((Verbose > 1) || (NumChannelsX[i] <= 0))
-	    Fprintf(stdout, "Number of x channels for layer %d is %d\n",
-				i, NumChannelsX[i]);
-	if ((Verbose > 1) || (NumChannelsY[i] <= 0))
-	    Fprintf(stdout, "Number of y channels for layer %d is %d\n",
-				i, NumChannelsY[i]);
-	
-	if (NumChannelsX[i] <= 0) {
-	    Fprintf(stderr, "Something wrong with layer %d x bounds.\n", i);
-	    return(-3);
-	}
-	if (NumChannelsY[i] <= 0) {
-	    Fprintf(stderr, "Something wrong with layer %d y bounds.\n", i);
-	    return(-3);
-	}
-	Flush(stdout);
+    if (PitchX == 0.0) {
+	Fprintf(stderr, "Have a 0 pitch for X direction.  Exit.\n");
+	return (-3);
     }
+    else if (PitchY == 0.0) {
+	Fprintf(stderr, "Have a 0 pitch for Y direction.  Exit.\n");
+	return (-3);
+    }
+
+    NumChannelsX = (int)(1.5 + (Xupperbound - Xlowerbound) / PitchX);
+    NumChannelsY = (int)(1.5 + (Yupperbound - Ylowerbound) / PitchY);
+    if ((Verbose > 1) || (NumChannelsX <= 0))
+	Fprintf(stdout, "Number of x channels is %d\n", NumChannelsX);
+    if ((Verbose > 1) || (NumChannelsY <= 0))
+	Fprintf(stdout, "Number of y channels is %d\n", NumChannelsY);
+	
+    if (NumChannelsX <= 0) {
+	Fprintf(stderr, "Something wrong with x bounds.\n");
+	return(-3);
+    }
+    if (NumChannelsY <= 0) {
+	Fprintf(stderr, "Something wrong with y bounds.\n");
+	return(-3);
+    }
+    Flush(stdout);
 
     // Go through all nodes and remove any tap or extend entries that are
     // out of bounds.
@@ -118,8 +151,8 @@ int set_num_channels(void)
 	    ltap = NULL;
 	    for (ctap = node->taps; ctap != NULL; ) {
 		ntap = ctap->next;
-		glimitx = NumChannelsX[ctap->layer];
-		glimity = NumChannelsY[ctap->layer];
+		glimitx = NumChannelsX;
+		glimity = NumChannelsY;
 		if (ctap->gridx < 0 || ctap->gridx >= glimitx ||
 				ctap->gridy < 0 || ctap->gridy >= glimity) {
 		    /* Remove ctap */
@@ -136,8 +169,8 @@ int set_num_channels(void)
 	    ltap = NULL;
 	    for (ctap = node->extend; ctap != NULL; ) {
 		ntap = ctap->next;
-		glimitx = NumChannelsX[ctap->layer];
-		glimity = NumChannelsY[ctap->layer];
+		glimitx = NumChannelsX;
+		glimity = NumChannelsY;
 		if (ctap->gridx < 0 || ctap->gridx >= glimitx ||
 				ctap->gridy < 0 || ctap->gridy >= glimity) {
 		    /* Remove ctap */
@@ -168,7 +201,7 @@ int allocate_obs_array(void)
    if (Obs[0] != NULL) return 0;	/* Already been called */
 
    for (i = 0; i < Num_layers; i++) {
-      Obs[i] = (u_int *)calloc(NumChannelsX[i] * NumChannelsY[i],
+      Obs[i] = (u_int *)calloc(NumChannelsX * NumChannelsY,
 			sizeof(u_int));
       if (!Obs[i]) {
 	 Fprintf(stderr, "Out of memory 4.\n");
@@ -366,34 +399,12 @@ runqrouter(int argc, char *argv[])
 #else
       fprintf(infoFILEptr, "qrouter %s.%s\n", VERSION, REVISION);
 #endif
+      /* Output database units expected by the technology LEF file */
+      /* Note that this comes from MANUFACTURINGGRID, not UNITS DATABASE */
+      fprintf(infoFILEptr, "units scale %d\n", Scales.mscale);
 
-      /* Resolve pitches.  This is normally done after reading	*/
-      /* the DEF file, but the info file is usually generated	*/
-      /* from LEF layer information only, in order to get the	*/
-      /* values needed to write the DEF file tracks.		*/
-
-      for (i = 0; i < Num_layers; i++) {
-	 int o = LefGetRouteOrientation(i);
-
-	 /* Set PitchX and PitchY from route info as	*/
-	 /* check_variable_pitch needs the values	*/
-
-	 if (o == 1)
-	    PitchY[i] = LefGetRoutePitch(i);
-	 else
-	    PitchX[i] = LefGetRoutePitch(i);
-      }
-
-      /* Resolve pitch information similarly to post_config() */
-
-      for (i = 1; i < Num_layers; i++) {
-	 int o = LefGetRouteOrientation(i);
-
-	 if ((o == 1) && (PitchY[i - 1] == 0))
-	    PitchY[i - 1] = PitchY[i];
-	 else if ((o == 0) && (PitchX[i - 1] == 0))
-	    PitchX[i - 1] = PitchX[i];
-      }
+      /* Resolve base horizontal and vertical pitches. */
+      post_config(TRUE);
 
       /* Print information about route layers, and exit */
       for (i = 0; i < Num_layers; i++) {
@@ -403,11 +414,9 @@ runqrouter(int argc, char *argv[])
 	 char *layername = LefGetRouteName(i);
 
 	 check_variable_pitch(i, &hnum, &vnum);
-	 if (vnum > 1 && hnum == 1) hnum++;	// see note in node.c
-	 if (hnum > 1 && vnum == 1) vnum++;
 		
 	 if (layername != NULL) {
-	    pitch = (o == 1) ? PitchY[i] : PitchX[i],
+	    pitch = (o == 1) ? PitchY : PitchX,
 	    width = LefGetRouteWidth(i);
 	    if (pitch == 0.0 || width == 0.0) continue;
 	    fprintf(infoFILEptr, "%s %g %g %g %s",
@@ -444,7 +453,7 @@ runqrouter(int argc, char *argv[])
    }
 
    Obs[0] = (u_int *)NULL;
-   NumChannelsX[0] = 0;	// This is so we can check if NumChannelsX/Y were
+   NumChannelsX = 0;	// This is so we can check if NumChannelsX/Y were
 			// set from within DefRead() due to reading in
 			// existing nets.
 
@@ -476,7 +485,7 @@ u_char remove_from_failed(NET net)
 	}
 	lastnl = nl;
     }
-    return FALSE;
+    return FALSE; 
 }
 
 /*--------------------------------------------------------------*/
@@ -495,6 +504,25 @@ void remove_failed()
     }
 }
 
+/*--------------------------------------------------------------*/
+/* Remove the first (top) route record from a net		*/
+/*--------------------------------------------------------------*/
+
+void remove_top_route(NET net)
+{
+    ROUTE rt;
+    SEG seg;
+
+    rt = net->routes;
+    net->routes = net->routes->next;
+    while (rt->segments) {
+	seg = rt->segments;
+	rt->segments = rt->segments->next;
+	    free(seg);
+    }
+    free(rt);
+}
+
 /*--------------------------------------------------------------*/
 /* reinitialize ---						*/
 /*								*/
@@ -516,7 +544,7 @@ static void reinitialize()
     // Free up all of the matrices
 
     for (i = 0; i < Pinlayers; i++) {
-	for (j = 0; j < NumChannelsX[i] * NumChannelsY[i]; j++)
+	for (j = 0; j < NumChannelsX * NumChannelsY; j++)
 	    if (Nodeinfo[i][j])
 		free(Nodeinfo[i][j]);
 	free(Nodeinfo[i]);
@@ -547,16 +575,9 @@ static void reinitialize()
 	    net->noripup = net->noripup->next;
 	    free(nl);
 	}
-	while (net->routes) {
-	    rt = net->routes;
-	    net->routes = net->routes->next;
-	    while (rt->segments) {
-		seg = rt->segments;
-		rt->segments = rt->segments->next;
-		free(seg);
-	    }
-	    free(rt);
-	}
+	while (net->routes)
+            remove_top_route(net);
+
 	while (net->netnodes) {
 	    node = net->netnodes;
 	    net->netnodes = net->netnodes->next;
@@ -608,6 +629,172 @@ static void reinitialize()
     Nlgates = NULL;
 }
 
+/*--------------------------------------------------------------*/
+/* apply_drc_blocks() ---					*/
+/*								*/
+/* Use via and route width and spacing information to determine	*/
+/* if blockages are needed in tracks adjacent to routed		*/
+/* segments to avoid causing DRC errors in the output.		*/
+/*								*/
+/* If layer == -1, then determine values normally for all	*/
+/* route layers.  If layer >= 0, determine values for specified	*/
+/* layer only.  If via_except > 0, then adjust the value for	*/
+/* a DRC violating distance for vias in adjacent tracks by that	*/
+/* amount (in microns).  If route_except > 0, then adjust the	*/
+/* value for a DRC violating distance between a via and a route	*/
+/* in adjacent tracks by that amount.				*/
+/*--------------------------------------------------------------*/
+
+void apply_drc_blocks(int layer, double via_except, double route_except)
+{
+   int i;
+   double sreq1, sreq2, sreq2t;
+
+   // Fill in needblock bit fields, which are used by commit_proute
+   // when route layers are too large for the grid size, and grid points
+   // around a route need to be marked as blocked whenever something is
+   // routed on those layers.
+
+   // "ROUTEBLOCK" is set if the spacing is violated between a normal
+   // route and an adjacent via.  "VIABLOCK" is set if the spacing is
+   // violated between two adjacent vias.  It may be helpful to define
+   // a third category which is route-to-route spacing violation.
+
+   // There are up to four different via types per base layer with
+   // different geometries based on the permutation of rotations of
+   // the top and bottom layers, so we only register blocking behavior
+   // if all of the via types will generate spacing violations.
+
+   for (i = 0; i < Num_layers; i++) {
+      if ((layer >= 0) && (i != layer)) continue;
+
+      needblock[i] = FALSE;
+
+      sreq1 = LefGetRouteSpacing(i);
+      if (i < Num_layers - 1) {
+         sreq2 = LefGetXYViaWidth(i, i, 0, 0) + sreq1;
+         sreq2t = LefGetXYViaWidth(i, i, 0, 1) + sreq1;
+         if (sreq2t < sreq2) sreq2 = sreq2t;
+         sreq2t = LefGetXYViaWidth(i, i, 0, 2) + sreq1;
+         if (sreq2t < sreq2) sreq2 = sreq2t;
+         sreq2t = LefGetXYViaWidth(i, i, 0, 3) + sreq1;
+         if (sreq2t < sreq2) sreq2 = sreq2t;
+	 sreq2 -= via_except;
+         if ((sreq2 - EPS) > PitchX) needblock[i] |= VIABLOCKX;
+      }
+      if (i != 0) {
+	 sreq2 = LefGetXYViaWidth(i - 1, i, 0, 0) + sreq1;
+	 sreq2t = LefGetXYViaWidth(i - 1, i, 0, 1) + sreq1;
+	 if (sreq2t < sreq2) sreq2 = sreq2t;
+	 sreq2t = LefGetXYViaWidth(i - 1, i, 0, 2) + sreq1;
+	 if (sreq2t < sreq2) sreq2 = sreq2t;
+	 sreq2t = LefGetXYViaWidth(i - 1, i, 0, 3) + sreq1;
+	 if (sreq2t < sreq2) sreq2 = sreq2t;
+	 sreq2 -= via_except;
+         if ((sreq2 - EPS) > PitchX) needblock[i] |= VIABLOCKX;
+      }
+
+      if (i < Num_layers - 1) {
+         sreq2 = LefGetXYViaWidth(i, i, 1, 0) + sreq1;
+         sreq2t = LefGetXYViaWidth(i, i, 1, 1) + sreq1;
+         if (sreq2t < sreq2) sreq2 = sreq2t;
+         sreq2t = LefGetXYViaWidth(i, i, 1, 2) + sreq1;
+         if (sreq2t < sreq2) sreq2 = sreq2t;
+         sreq2t = LefGetXYViaWidth(i, i, 1, 3) + sreq1;
+         if (sreq2t < sreq2) sreq2 = sreq2t;
+	 sreq2 -= via_except;
+         if ((sreq2 - EPS) > PitchY) needblock[i] |= VIABLOCKY;
+      }
+      if (i != 0) {
+	 sreq2 = LefGetXYViaWidth(i - 1, i, 1, 0) + sreq1;
+	 sreq2t = LefGetXYViaWidth(i - 1, i, 1, 1) + sreq1;
+	 if (sreq2t < sreq2) sreq2 = sreq2t;
+	 sreq2t = LefGetXYViaWidth(i - 1, i, 1, 2) + sreq1;
+	 if (sreq2t < sreq2) sreq2 = sreq2t;
+	 sreq2t = LefGetXYViaWidth(i - 1, i, 1, 3) + sreq1;
+	 if (sreq2t < sreq2) sreq2 = sreq2t;
+	 sreq2 -= via_except;
+         if ((sreq2 - EPS) > PitchY) needblock[i] |= VIABLOCKY;
+      }
+
+      sreq1 += 0.5 * LefGetRouteWidth(i);
+
+      if (i < Num_layers - 1) {
+         sreq2 = sreq1 + 0.5 * LefGetXYViaWidth(i, i, 0, 0);
+         sreq2t = sreq1 + 0.5 * LefGetXYViaWidth(i, i, 0, 1);
+         if (sreq2t < sreq2) sreq2 = sreq2t;
+         sreq2t = sreq1 + 0.5 * LefGetXYViaWidth(i, i, 0, 2);
+         if (sreq2t < sreq2) sreq2 = sreq2t;
+         sreq2t = sreq1 + 0.5 * LefGetXYViaWidth(i, i, 0, 3);
+         if (sreq2t < sreq2) sreq2 = sreq2t;
+	 sreq2 -= route_except;
+         if ((sreq2 - EPS) > PitchX) needblock[i] |= ROUTEBLOCKX;
+      }
+      if (i != 0) {
+	 sreq2 = sreq1 + 0.5 * LefGetXYViaWidth(i - 1, i, 0, 0);
+	 sreq2t = sreq1 + 0.5 * LefGetXYViaWidth(i - 1, i, 0, 1);
+	 if (sreq2t < sreq2) sreq2 = sreq2t;
+	 sreq2t = sreq1 + 0.5 * LefGetXYViaWidth(i - 1, i, 0, 2);
+	 if (sreq2t < sreq2) sreq2 = sreq2t;
+	 sreq2t = sreq1 + 0.5 * LefGetXYViaWidth(i - 1, i, 0, 3);
+	 if (sreq2t < sreq2) sreq2 = sreq2t;
+	 sreq2 -= route_except;
+         if ((sreq2 - EPS) > PitchX) needblock[i] |= ROUTEBLOCKX;
+      }
+
+      if (i < Num_layers - 1) {
+         sreq2 = sreq1 + 0.5 * LefGetXYViaWidth(i, i, 1, 0);
+         sreq2t = sreq1 + 0.5 * LefGetXYViaWidth(i, i, 1, 1);
+         if (sreq2t < sreq2) sreq2 = sreq2t;
+         sreq2t = sreq1 + 0.5 * LefGetXYViaWidth(i, i, 1, 2);
+         if (sreq2t < sreq2) sreq2 = sreq2t;
+         sreq2t = sreq1 + 0.5 * LefGetXYViaWidth(i, i, 1, 3);
+         if (sreq2t < sreq2) sreq2 = sreq2t;
+	 sreq2 -= route_except;
+         if ((sreq2 - EPS) > PitchY) needblock[i] |= ROUTEBLOCKY;
+      }
+      if (i != 0) {
+	 sreq2 = sreq1 + 0.5 * LefGetXYViaWidth(i - 1, i, 1, 0);
+	 sreq2t = sreq1 + 0.5 * LefGetXYViaWidth(i - 1, i, 1, 1);
+	 if (sreq2t < sreq2) sreq2 = sreq2t;
+	 sreq2t = sreq1 + 0.5 * LefGetXYViaWidth(i - 1, i, 1, 2);
+	 if (sreq2t < sreq2) sreq2 = sreq2t;
+	 sreq2t = sreq1 + 0.5 * LefGetXYViaWidth(i - 1, i, 1, 3);
+	 if (sreq2t < sreq2) sreq2 = sreq2t;
+	 sreq2 -= route_except;
+         if ((sreq2 - EPS) > PitchY) needblock[i] |= ROUTEBLOCKY;
+      }
+   }
+}
+
+/*--------------------------------------------------------------*/
+/* remove_tap_blocks						*/
+/*								*/
+/* Qrouter avoids routing directly over a tap point, blocking	*/
+/* it, if there is a Nodeinfo[][]->nodeloc entry present.	*/
+/* Remove this entry to remove the blockage (it can be		*/
+/* replaced if needed by copying back the Nodeinfo[][]->nodesav	*/
+/* pointer)							*/
+/*--------------------------------------------------------------*/
+
+void
+remove_tap_blocks(int netnum)
+{
+    int i, j;
+    NODE node;
+
+    for (i = 0; i < Pinlayers; i++) {
+	for (j = 0; j < NumChannelsX * NumChannelsY; j++) {
+	    if (Nodeinfo[i][j]) {
+		node = Nodeinfo[i][j]->nodeloc;
+		if (node != (NODE)NULL)
+		    if (node->netnum == netnum)
+			Nodeinfo[i][j]->nodeloc = (NODE)NULL;
+	    }
+        }
+    }
+}
+
 /*--------------------------------------------------------------*/
 /* post_def_setup ---						*/
 /*								*/
@@ -618,8 +805,9 @@ static void reinitialize()
 static int post_def_setup()
 {
    NET net;
+   ROUTE rt;
+   DPOINT tpoint;
    int i;
-   double sreq1, sreq2;
 
    if (DEFfilename == NULL) {
       Fprintf(stderr, "No DEF file read, nothing to set up.\n");
@@ -647,14 +835,14 @@ static int post_def_setup()
 
    for (i = 0; i < Num_layers; i++) {
 
-      Obsinfo[i] = (float *)calloc(NumChannelsX[i] * NumChannelsY[i],
+      Obsinfo[i] = (float *)calloc(NumChannelsX * NumChannelsY,
 			sizeof(float));
       if (!Obsinfo[i]) {
 	 fprintf(stderr, "Out of memory 5.\n");
 	 exit(5);
       }
 
-      Nodeinfo[i] = (NODEINFO *)calloc(NumChannelsX[i] * NumChannelsY[i],
+      Nodeinfo[i] = (NODEINFO *)calloc(NumChannelsX * NumChannelsY,
 			sizeof(NODEINFO));
       if (!Nodeinfo[i]) {
 	 fprintf( stderr, "Out of memory 6.\n");
@@ -665,12 +853,32 @@ static int post_def_setup()
 
    if (Verbose > 1)
       Fprintf(stderr, "Diagnostic: memory block is %d bytes\n",
-		(int)sizeof(u_int) * NumChannelsX[0] * NumChannelsY[0]);
+		(int)sizeof(u_int) * NumChannelsX * NumChannelsY);
+
+   /* If any watch points were made, make sure that they have	*/
+   /* the correct geometry values, since they were made before	*/
+   /* the DEF file was read.					*/
+
+   for (tpoint = testpoint; tpoint; tpoint = tpoint->next) {
+	if (tpoint->gridx < 0) {
+	    /* Compute gridx,y from x,y */
+	    tpoint->gridx = (int)(round((tpoint->x - Xlowerbound) / PitchX));
+	    tpoint->gridy = (int)(round((tpoint->y - Xlowerbound) / PitchX));
+	}
+	else {
+	    /* Compute x,y from gridx,y */
+	    tpoint->x = (tpoint->gridx * PitchX) + Xlowerbound;
+	    tpoint->y = (tpoint->gridy * PitchY) + Ylowerbound;
+	}
+   }
 
    /* Be sure to create obstructions from gates first, since we don't	*/
    /* want improperly defined or positioned obstruction layers to over-	*/
    /* write our node list.						*/
 
+#ifdef TCL_QROUTER
+   find_free_antenna_taps(antenna_cell);
+#endif
    expand_tap_geometry();
    clip_gate_taps();
    create_obstructions_from_gates();
@@ -680,13 +888,16 @@ static int post_def_setup()
    create_obstructions_from_variable_pitch();
    adjust_stub_lengths();
    find_route_blocks();
-   count_reachable_taps();
+   count_reachable_taps(unblockAll);
    count_pinlayers();
    
-   // If any nets are pre-routed, place those routes.
+   // If any nets are pre-routed, calculate route endpoints, and
+   // place those routes.
 
    for (i = 0; i < Numnets; i++) {
       net = Nlnets[i];
+      for (rt = net->routes; rt; rt = rt->next)
+	 route_set_connections(net, rt);
       writeback_all_routes(net);
    }
 
@@ -696,7 +907,7 @@ static int post_def_setup()
    for (i = 0; i < Num_layers; i++) free(Obsinfo[i]);
 
    for (i = 0; i < Num_layers; i++) {
-      Obs2[i] = (PROUTE *)calloc(NumChannelsX[i] * NumChannelsY[i],
+      Obs2[i] = (PROUTE *)calloc(NumChannelsX * NumChannelsY,
 			sizeof(PROUTE));
       if (!Obs2[i]) {
          fprintf( stderr, "Out of memory 9.\n");
@@ -704,50 +915,13 @@ static int post_def_setup()
       }
    }
 
-   // Fill in needblock bit fields, which are used by commit_proute
-   // when route layers are too large for the grid size, and grid points
-   // around a route need to be marked as blocked whenever something is
-   // routed on those layers.
+   // Remove tap blocks from power, ground, and antenna nets, as these
+   // can take up large areas of the layout and will cause serious issues
+   // with routability if left blocked.
 
-   // "ROUTEBLOCK" is set if the spacing is violated between a normal
-   // route and an adjacent via.  "VIABLOCK" is set if the spacing is
-   // violated between two adjacent vias.  It may be helpful to define
-   // a third category which is route-to-route spacing violation.
-
-   for (i = 0; i < Num_layers; i++) {
-      needblock[i] = FALSE;
-      sreq1 = LefGetRouteSpacing(i);
-
-      sreq2 = LefGetViaWidth(i, i, 0) + sreq1;
-      if ((sreq2 - EPS) > PitchX[i]) needblock[i] |= VIABLOCKX;
-      if (i != 0) {
-	 sreq2 = LefGetViaWidth(i - 1, i, 0) + sreq1;
-         if ((sreq2 - EPS) > PitchX[i]) needblock[i] |= VIABLOCKX;
-      }
-
-      sreq2 = LefGetViaWidth(i, i, 1) + sreq1;
-      if ((sreq2 - EPS) > PitchY[i]) needblock[i] |= VIABLOCKY;
-      if (i != 0) {
-	 sreq2 = LefGetViaWidth(i - 1, i, 1) + sreq1;
-         if ((sreq2 - EPS) > PitchY[i]) needblock[i] |= VIABLOCKY;
-      }
-
-      sreq1 += 0.5 * LefGetRouteWidth(i);
-
-      sreq2 = sreq1 + 0.5 * LefGetViaWidth(i, i, 0);
-      if ((sreq2 - EPS) > PitchX[i]) needblock[i] |= ROUTEBLOCKX;
-      if (i != 0) {
-	 sreq2 = sreq1 + 0.5 * LefGetViaWidth(i - 1, i, 0);
-         if ((sreq2 - EPS) > PitchX[i]) needblock[i] |= ROUTEBLOCKX;
-      }
-
-      sreq2 = sreq1 + 0.5 * LefGetViaWidth(i, i, 1);
-      if ((sreq2 - EPS) > PitchY[i]) needblock[i] |= ROUTEBLOCKY;
-      if (i != 0) {
-	 sreq2 = sreq1 + 0.5 * LefGetViaWidth(i - 1, i, 1);
-         if ((sreq2 - EPS) > PitchY[i]) needblock[i] |= ROUTEBLOCKY;
-      }
-   }
+   remove_tap_blocks(VDD_NET);
+   remove_tap_blocks(GND_NET);
+   remove_tap_blocks(ANTENNA_NET);
 
    // Now we have netlist data, and can use it to get a list of nets.
 
@@ -763,15 +937,18 @@ static int post_def_setup()
 /* read_def ---							*/
 /*								*/
 /* Read in the DEF file in DEFfilename				*/
+/* Return 0 on success, 1 on fatal error in DEF file.		*/
 /*--------------------------------------------------------------*/
 
-void read_def(char *filename)
+int read_def(char *filename)
 {
-   double oscale, precis;
+   float oscale;
+   double precis;
+   int result;
 
    if ((filename == NULL) && (DEFfilename == NULL)) {
       Fprintf(stderr, "No DEF file specified, nothing to read.\n");
-      return;
+      return 1;
    }
    else if (filename != NULL) {
       if (DEFfilename != NULL) {
@@ -782,8 +959,9 @@ void read_def(char *filename)
    }
    else reinitialize();
 
-   oscale = (double)DefRead(DEFfilename);
-   precis = Scales.mscale / oscale;	// from LEF manufacturing grid
+   oscale = (float)0.0;
+   result = DefRead(DEFfilename, &oscale);
+   precis = Scales.mscale / (double)oscale;	// from LEF manufacturing grid
    if (precis < 1.0) precis = 1.0;
    precis *= (double)Scales.iscale;	// user-defined extra scaling
 
@@ -796,6 +974,7 @@ void read_def(char *filename)
 		1.0 / (double)Scales.iscale);
 
    post_def_setup();
+   return result;
 }
 
 /*--------------------------------------------------------------*/
@@ -826,6 +1005,7 @@ int dofirststage(u_char graphdebug, int debug_netnum)
 	    if (Verbose > 0)
 	       Fprintf(stdout, "Finished routing net %s\n", net->netname);
 	    Fprintf(stdout, "Nets remaining: %d\n", remaining);
+	    Flush(stdout);
 	 }
 	 else {
 	    if (Verbose > 0)
@@ -886,7 +1066,7 @@ NET getnettoroute(int order)
    // only one node.
 
    if ((net->numnodes == 1) && (net->netnum == VDD_NET ||
-		net->netnum == GND_NET))
+		net->netnum == GND_NET || net->netnum == ANTENNA_NET))
       return net;
 
    if (Verbose > 3) {
@@ -1283,12 +1463,15 @@ int dothirdstage(u_char graphdebug, int debug_netnum, u_int effort)
 	       Fprintf(stdout, "Finished routing net %s\n", net->netname);
 	    remaining--;
 	    Fprintf(stdout, "Nets remaining: %d\n", remaining);
+	    Flush(stdout);
 	    remove_routes(rt, FALSE);	/* original is no longer needed */
 	 }
 	 else if (!failed) {
 	    if (Verbose > 0)
 	       Fprintf(stdout, "Failed to route net %s; restoring original\n",
 			net->netname);
+
+	    ripup_net(net, TRUE, FALSE, TRUE);	// Remove routes from Obs array
 	    remove_routes(net->routes, FALSE);	/* should be NULL already */
 	    net->routes = rt;
 	    writeback_all_routes(net);	/* restore the original */
@@ -1579,18 +1762,7 @@ static int next_route_setup(struct routeinfo_ *iroute, u_char stage)
 
      // Remove nodes of the net from Nodeinfo.nodeloc so that they will not be
      // used for crossover costing of future routes.
-
-     for (i = 0; i < Pinlayers; i++) {
-	for (j = 0; j < NumChannelsX[i] * NumChannelsY[i]; j++) {
-	   if (Nodeinfo[i][j]) {
-	      node = Nodeinfo[i][j]->nodeloc;
-	      if (node != (NODE)NULL)
-	         if (node->netnum == iroute->net->netnum)
-		    Nodeinfo[i][j]->nodeloc = (NODE)NULL;
-	   }
-        }
-     }
-
+     remove_tap_blocks(iroute->net->netnum);
      free_glist(iroute);
      return 0;
   }
@@ -1637,13 +1809,13 @@ static int route_setup(struct routeinfo_ *iroute, u_char stage)
   // terminal positions for the net being routed.
 
   for (i = 0; i < Num_layers; i++) {
-      for (j = 0; j < NumChannelsX[i] * NumChannelsY[i]; j++) {
+      for (j = 0; j < NumChannelsX * NumChannelsY; j++) {
 	  netnum = Obs[i][j] & (~BLOCKED_MASK);
 	  Pr = &Obs2[i][j];
 	  if (netnum != 0) {
 	      Pr->flags = 0;		// Clear all flags
-	      if (netnum == DRC_BLOCKAGE)
-	         Pr->prdata.net = netnum;
+	      if ((netnum & DRC_BLOCKAGE) == DRC_BLOCKAGE)
+	         Pr->prdata.net = DRC_BLOCKAGE;
 	      else
 	         Pr->prdata.net = netnum & NETNUM_MASK;
 	   } else {
@@ -1653,7 +1825,8 @@ static int route_setup(struct routeinfo_ *iroute, u_char stage)
       }
   }
 
-  if (iroute->net->netnum == VDD_NET || iroute->net->netnum == GND_NET) {
+  if (iroute->net->netnum == VDD_NET || iroute->net->netnum == GND_NET ||
+		iroute->net->netnum == ANTENNA_NET) {
      // The normal method of selecting source and target is not amenable
      // to power bus routes.  Instead, we use the global standard cell
      // power rails as the target, and each net in sequence becomes the
@@ -1682,8 +1855,8 @@ static int route_setup(struct routeinfo_ *iroute, u_char stage)
 
   if (result) {
      iroute->bbox.x2 = iroute->bbox.y2 = 0;
-     iroute->bbox.x1 = NumChannelsX[0];
-     iroute->bbox.y1 = NumChannelsY[0];
+     iroute->bbox.x1 = NumChannelsX;
+     iroute->bbox.y1 = NumChannelsY;
 
      if (iroute->do_pwrbus == FALSE) {
 
@@ -1758,18 +1931,7 @@ static int route_setup(struct routeinfo_ *iroute, u_char stage)
   if (!result) {
      // Remove nodes of the net from Nodeinfo.nodeloc so that they will not be
      // used for crossover costing of future routes.
-
-     for (i = 0; i < Pinlayers; i++) {
-	for (j = 0; j < NumChannelsX[i] * NumChannelsY[i]; j++) {
-	   if (Nodeinfo[i][j]) {
-	      iroute->nsrc = Nodeinfo[i][j]->nodeloc;
-	      if (iroute->nsrc != (NODE)NULL)
-	         if (iroute->nsrc->netnum == iroute->net->netnum)
-		    Nodeinfo[i][j]->nodeloc = (NODE)NULL;
-	   }
-        }
-     }
-
+     remove_tap_blocks(iroute->net->netnum);
      free_glist(iroute);
      return 0;
   }
@@ -1844,7 +2006,7 @@ static int route_setup(struct routeinfo_ *iroute, u_char stage)
 /*   AUTHOR and DATE: steve beccue      Fri Aug 8		*/
 /*--------------------------------------------------------------*/
 
-static int route_segs(struct routeinfo_ *iroute, u_char stage, u_char graphdebug)
+int route_segs(struct routeinfo_ *iroute, u_char stage, u_char graphdebug)
 {
   POINT gpoint, gunproc, newpt;
   int  i, o;
@@ -1895,6 +2057,18 @@ static int route_segs(struct routeinfo_ *iroute, u_char stage, u_char graphdebug
 
       iroute->glist[0] = gpoint->next;
 
+      // Stop-gap:  Needs to be investigated.  Occasional gpoint has
+      // large (random?) value for y1.  Suggests a memory leak.  Only
+      // seen occurring during doantennaroute().  Check using valgrind.
+
+      if ((gpoint->x1 < 0) || (gpoint->y1 < 0) ||
+		(gpoint->x1 > NumChannelsX) ||
+		(gpoint->y1 > NumChannelsY)) {
+         Fprintf(stderr, "Internal memory error!\n");
+	 freePOINT(gpoint);
+	 continue;
+      }
+
       curpt.x = gpoint->x1;
       curpt.y = gpoint->y1;
       curpt.lay = gpoint->layer;
@@ -2016,7 +2190,7 @@ static int route_segs(struct routeinfo_ *iroute, u_char stage, u_char graphdebug
 	       predecessor = PR_CONFLICT;
 	    case EAST:
 	       predecessor |= PR_PRED_W;
-               if ((curpt.x + 1) < NumChannelsX[curpt.lay]) {
+               if ((curpt.x + 1) < NumChannelsX) {
          	  if ((gpoint = eval_pt(&curpt, predecessor, stage)) != NULL) {
          	     gpoint->next = iroute->glist[i];
          	     iroute->glist[i] = gpoint;
@@ -2052,7 +2226,7 @@ static int route_segs(struct routeinfo_ *iroute, u_char stage, u_char graphdebug
 	       predecessor = PR_CONFLICT;
 	    case NORTH:
 	       predecessor |= PR_PRED_S;
-               if ((curpt.y + 1) < NumChannelsY[curpt.lay]) {
+               if ((curpt.y + 1) < NumChannelsY) {
          	  if ((gpoint = eval_pt(&curpt, predecessor, stage)) != NULL) {
          	     gpoint->next = iroute->glist[i];
          	     iroute->glist[i] = gpoint;
@@ -2162,7 +2336,7 @@ done:
 /*   AUTHOR and DATE: steve beccue      Fri Aug 8		*/
 /*--------------------------------------------------------------*/
 
-static ROUTE createemptyroute(void)
+ROUTE createemptyroute(void)
 {
    ROUTE rt;
 
