@@ -7,7 +7,7 @@
 
 #ifndef QROUTER_H
 
-#define OGRID(x, y, layer) ((int)((x) + ((y) * NumChannelsX[(layer)])))
+#define OGRID(x, y) ((int)((x) + ((y) * NumChannelsX)))
 #define MIN(x, y) (((x) < (y)) ? (x) : (y))
 #define MAX(x, y) (((x) > (y)) ? (x) : (y))
 #define ABSDIFF(x, y) (((x) > (y)) ? ((x) - (y)) : ((y) - (x)))
@@ -40,12 +40,16 @@ typedef unsigned long  u_long;
 #endif /* _SYS_TYPES_H */
 
 /* Compare functions aren't defined in the Mac's standard library */
-#if defined(__APPLE__)
+#if defined(__APPLE__) || defined(__FreeBSD__)
 typedef int (*__compar_fn_t)(const void*, const void*);
 #endif
 
 /* Maximum number of route layers */
-#define MAX_LAYERS    9
+#define MAX_LAYERS    12
+
+/* Maximum number of all defined layers.  Since masterslice and	*/
+/* overlap types are ignored, this just includes all the cuts.	*/
+#define MAX_TYPES    (MAX_LAYERS * 2 - 1)
 
 /* Cell name (and other names) max length */
 #define MAX_NAME_LEN    1024
@@ -59,8 +63,9 @@ typedef int (*__compar_fn_t)(const void*, const void*);
 // define possible gate orientations
 
 #define MNONE	0
-#define MX	1
-#define MY	2
+#define MX	1   // flipped in X
+#define MY	2   // flipped in Y
+#define R90	4   // rotated clockwise 90 degrees
 
 // define search directions
 
@@ -151,6 +156,7 @@ struct string_ {
 #define	 ST_OFFSET_START	0x04	/* (x1, y1) is offset from grid */
 #define	 ST_OFFSET_END		0x08	/* (x2, y2) is offset from grid */
 #define  ST_SPECIAL		0x10	/* wide metal (special net)	*/
+#define  ST_MINMETAL		0x20	/* segment for min metal area	*/
 
 typedef struct seg_ *SEG;
 
@@ -223,6 +229,7 @@ struct route_ {
 #define RT_END_NODE	0x08	// Route ends on a node
 #define RT_VISITED	0x10	// Flag for recursive search
 #define RT_RIP		0x20	// Flag for route rip-up
+#define RT_CHECK	0x40	// Route from DEF file needs checking
 
 /* Structure used to hold nodes, saved nodes, and stub/offset info */
 
@@ -246,6 +253,8 @@ struct nodeinfo_ {
 #define NI_OFFSET_MASK   0x0c	// Tap offset mask (N/S + E/W)
 #define NI_NO_VIAX   	 0x10	// Via in ViaX array is prohibited
 #define NI_NO_VIAY   	 0x20	// Via in ViaY array is prohibited
+#define NI_VIA_X 	 0x40	// Placed via is oriented horizontally
+#define NI_VIA_Y 	 0x80	// Placed via is oriented vertically
 
 struct node_ {
   NODE    next;
@@ -274,6 +283,7 @@ struct gate_ {
     char **node;	// names of the pins on this gate
     int   *netnum;	// net number connected to each pin
     NODE  *noderec;	// node record for each pin
+    float *area;	// gate area for each pin
     u_char *direction;	// port direction (input, output, etc.)	
     DSEG  *taps;	// list of gate node locations and layers
     DSEG   obs;		// list of obstructions in gate
@@ -283,6 +293,16 @@ struct gate_ {
     int orient;
 };
 
+// Define record holding information pointing to a gate and the
+// index into a specific node of that gate.
+
+typedef struct gatenode_ *GATENODE;
+
+struct gatenode_ {
+    GATE gate;
+    int idx;
+};
+
 // Structure for a network to be routed
 
 typedef struct net_ *NET;
@@ -415,6 +435,7 @@ extern NETLIST FailedNets;	// nets that have failed the first pass
 extern char    *DEFfilename;
 extern char    *delayfilename;
 extern ScaleRec Scales;
+extern DPOINT	testpoint;	// for debugging routing problems
 
 extern GATE   GateInfo;		// standard cell macro information
 extern GATE   PinMacro;		// macro definition for a pin
@@ -428,13 +449,13 @@ extern float  *Obsinfo[MAX_LAYERS];	// temporary detailed obstruction info
 extern NODEINFO *Nodeinfo[MAX_LAYERS];	// stub route distances to pins and
 					// pointers to node structures.
 
-#define NODEIPTR(x, y, l) (Nodeinfo[l][OGRID(x, y, l)])
-#define OBSINFO(x, y, l) (Obsinfo[l][OGRID(x, y, l)])
-#define OBSVAL(x, y, l)  (Obs[l][OGRID(x, y, l)])
-#define OBS2VAL(x, y, l) (Obs2[l][OGRID(x, y, l)])
+#define NODEIPTR(x, y, l) (Nodeinfo[l][OGRID(x, y)])
+#define OBSINFO(x, y, l) (Obsinfo[l][OGRID(x, y)])
+#define OBSVAL(x, y, l)  (Obs[l][OGRID(x, y)])
+#define OBS2VAL(x, y, l) (Obs2[l][OGRID(x, y)])
 
-#define RMASK(x, y)      (RMask[OGRID(x, y, 0)])
-#define CONGEST(x, y)	 (Congestion[OGRID(x, y, 0)])
+#define RMASK(x, y)      (RMask[OGRID(x, y)])
+#define CONGEST(x, y)	 (Congestion[OGRID(x, y)])
 
 extern DSEG  UserObs;			// user-defined obstruction layers
 
@@ -448,9 +469,11 @@ extern u_char forceRoutable;
 extern u_char maskMode;
 extern u_char mapType;
 extern u_char ripLimit;
+extern u_char unblockAll;
 
 extern char *vddnet;
 extern char *gndnet;
+extern char *antenna_cell;
 
 /* Tcl output to console handling */
 
@@ -466,18 +489,24 @@ extern char *gndnet;
 
 /* Function prototypes */
 
+void update_mscale(int mscale);
 static int next_route_setup(struct routeinfo_ *iroute, u_char stage);
 static int route_setup(struct routeinfo_ *iroute, u_char stage);
-static int route_segs(struct routeinfo_ *iroute, u_char stage, u_char graphdebug);
-static ROUTE createemptyroute(void);
+int route_segs(struct routeinfo_ *iroute, u_char stage, u_char graphdebug);
+ROUTE createemptyroute(void);
 static void helpmessage(void);
 
 int    set_num_channels(void);
 int    allocate_obs_array(void);
 int    countlist(NETLIST net);
 int    runqrouter(int argc, char *argv[]);
+void   remove_failed();
+void   apply_drc_blocks(int, double, double);
+void   remove_top_route(NET net);
+char  *get_annotate_info(NET net, char **pinptr);
 
-void   read_def(char *filename);
+
+int    read_def(char *filename);
 
 #ifdef TCL_QROUTER
 int    write_delays(char *filename);
