@@ -11,6 +11,7 @@
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <errno.h>
+#include <math.h>	/* for round() */
 
 #include <tk.h>
 
@@ -21,6 +22,7 @@
 #include "maze.h"
 #include "qconfig.h"
 #include "lef.h"
+#include "def.h"
 #include "graphics.h"
 #include "node.h"
 #include "tkSimple.h"
@@ -59,12 +61,21 @@ static int qrouter_stage2(
 static int qrouter_stage3(
     ClientData clientData, Tcl_Interp *interp,
     int objc, Tcl_Obj *CONST objv[]);
+static int qrouter_cleanup(
+    ClientData clientData, Tcl_Interp *interp,
+    int objc, Tcl_Obj *CONST objv[]);
 static int qrouter_writedef(
     ClientData clientData, Tcl_Interp *interp,
     int objc, Tcl_Obj *CONST objv[]);
+static int qrouter_writefailed(
+    ClientData clientData, Tcl_Interp *interp,
+    int objc, Tcl_Obj *CONST objv[]);
 static int qrouter_writedelays(
     ClientData clientData, Tcl_Interp *interp,
     int objc, Tcl_Obj *CONST objv[]);
+static int qrouter_antenna(
+    ClientData clientData, Tcl_Interp *interp,
+    int objc, Tcl_Obj *CONST objv[]);
 static int qrouter_readdef(
     ClientData clientData, Tcl_Interp *interp,
     int objc, Tcl_Obj *CONST objv[]);
@@ -110,6 +121,12 @@ static int qrouter_congested(
 static int qrouter_layers(
     ClientData clientData, Tcl_Interp *interp,
     int objc, Tcl_Obj *CONST objv[]);
+static int qrouter_drc(
+    ClientData clientData, Tcl_Interp *interp,
+    int objc, Tcl_Obj *CONST objv[]);
+static int qrouter_query(
+    ClientData clientData, Tcl_Interp *interp,
+    int objc, Tcl_Obj *CONST objv[]);
 static int qrouter_passes(
     ClientData clientData, Tcl_Interp *interp,
     int objc, Tcl_Obj *CONST objv[]);
@@ -128,6 +145,15 @@ static int qrouter_print(
 static int qrouter_quit(
     ClientData clientData, Tcl_Interp *interp,
     int objc, Tcl_Obj *CONST objv[]);
+static int qrouter_pitchx(
+    ClientData clientData, Tcl_Interp *interp,
+    int objc, Tcl_Obj *CONST objv[]);
+static int qrouter_pitchy(
+    ClientData clientData, Tcl_Interp *interp,
+    int objc, Tcl_Obj *CONST objv[]);
+static int qrouter_unblock(
+    ClientData clientData, Tcl_Interp *interp,
+    int objc, Tcl_Obj *CONST objv[]);
 
 static cmdstruct qrouter_commands[] =
 {
@@ -136,20 +162,28 @@ static cmdstruct qrouter_commands[] =
    {"stage1", qrouter_stage1},
    {"stage2", qrouter_stage2},
    {"stage3", qrouter_stage3},
+   {"cleanup", qrouter_cleanup},
    {"write_def", qrouter_writedef},
    {"read_def", qrouter_readdef},
    {"read_lef", qrouter_readlef},
    {"read_config", qrouter_readconfig},
    {"write_delays", qrouter_writedelays},
+   {"antenna", qrouter_antenna},
+   {"write_failed", qrouter_writefailed},
    {"layer_info", qrouter_layerinfo},
    {"obstruction", qrouter_obs},
    {"ignore", qrouter_ignore},
    {"priority", qrouter_priority},
+   {"pitchx", qrouter_pitchx},
+   {"pitchy", qrouter_pitchy},
+   {"unblock", qrouter_unblock},
    {"via", qrouter_via},
    {"resolution", qrouter_resolution},
    {"congested", qrouter_congested},
    {"layers", qrouter_layers},
+   {"drc", qrouter_drc},
    {"passes", qrouter_passes},
+   {"query", qrouter_query},
    {"vdd", qrouter_vdd},
    {"gnd", qrouter_gnd},
    {"failing", qrouter_failing},
@@ -297,7 +331,7 @@ void tcl_stdflush(FILE *f)
    char *stdptr = stdstr + 11;
     
    Tcl_SaveResult(qrouterinterp, &state);
-   strcpy(stdptr, (f == stderr) ? "err" : "out");
+   strncpy(stdptr, (f == stderr) ? "err" : "out", 3);
    Tcl_Eval(qrouterinterp, stdstr);
    Tcl_RestoreResult(qrouterinterp, &state);
 }
@@ -603,6 +637,7 @@ qrouter_start(ClientData clientData, Tcl_Interp *interp,
 	argv[argc++] = strdup(Tcl_GetString(objv[i]));
     }
 
+    init_config();
     result = runqrouter(argc, argv);
     if ((result == 0) && (batchmode == 0)) GUI_init(interp);
 
@@ -626,18 +661,30 @@ qrouter_start(ClientData clientData, Tcl_Interp *interp,
 	    fclose(scriptf);
 	    result = Tcl_EvalFile(interp, scriptfile);
 	}
-	free(scriptfile);
 
 	/* The script file should determine whether or not to	*/
 	/* exit by including the "quit" command.  But if there	*/
 	/* is an error in the script, then always quit.		*/
 
+	/* If tkcon console is in use and there is an error in	*/
+	/* the script, then print the error message to the	*/
+	/* terminal, not the console, or else it vanishes.	*/
+
 	if (result != TCL_OK) {
+	    if (consoleinterp == interp)
+		Fprintf(stderr, "Script file \"%s\" failed with result \'%s\'\n",
+			scriptfile, Tcl_GetStringResult(interp));
+	    else
+		fprintf(stderr, "Script file \"%s\" failed with result \'%s\'\n",
+			scriptfile, Tcl_GetStringResult(interp));
+	    free(scriptfile);
 	    /* Make sure Tcl has generated all output */
 	    while (Tcl_DoOneEvent(TCL_DONT_WAIT) != 0);
 	    /* And exit gracefully */
 	    qrouter_quit(clientData, interp, 1, objv);
 	}
+	else
+	    free(scriptfile);
     }
 
     if ((DEFfilename != NULL) && (Nlgates == NULL)) {
@@ -664,6 +711,11 @@ qrouter_quit(ClientData clientData, Tcl_Interp *interp,
 	return TCL_ERROR;
     }
 
+    /* Free up failed net list */
+    remove_failed();
+
+    /* Should be doing other cleanup tasks here. . . */
+
     if (consoleinterp == interp)
 	Tcl_Exit(TCL_OK);
     else
@@ -757,29 +809,13 @@ qrouter_map(ClientData clientData, Tcl_Interp *interp,
     return QrouterTagCallback(interp, objc, objv);
 }
 
-/*------------------------------------------------------*/
-/* Find the net named "netname" in the list of nets	*/
-/* and return a pointer to it.				*/
-/*							*/
-/* NOTE:  Really need a hash table lookup here!		*/
-/*------------------------------------------------------*/
-
-NET LookupNet(char *netname)
-{
-    NET net;
-    int i;
-
-    for (i = 0; i < Numnets; i++) {
-       net = Nlnets[i];
-       if (!strcmp(net->netname, netname))
-	  return net;
-    }
-    return NULL;
-}
-
 /*------------------------------------------------------*/
 /* Find the net with number "number" in the list of	*/
 /* nets and return a pointer to it.			*/
+/*							*/
+/* NOTE:  This could be hashed like net names, but is	*/
+/* only used in one place, and router performance does	*/
+/* not depend on it.					*/
 /*------------------------------------------------------*/
 
 NET LookupNetNr(int number)
@@ -795,22 +831,6 @@ NET LookupNetNr(int number)
     return NULL;
 }
 
-/*------------------------------------------------------*/
-/* Find the gate instance named gatename and return a	*/
-/* pointer to it.					*/
-/*------------------------------------------------------*/
-
-GATE LookupGate(char *gatename)
-{
-    GATE gate;
-
-    for (gate = Nlgates; gate != NULL; gate = gate->next) {
-       if (!strcmp(gate->gatename, gatename))
-	  return gate;
-    }
-    return NULL;
-}
-
 /*------------------------------------------------------*/
 /* Command "stage1"					*/
 /*							*/
@@ -899,7 +919,7 @@ qrouter_stage1(ClientData clientData, Tcl_Interp *interp,
 			return TCL_ERROR;
 		    }
 		    i++;
-		    net = LookupNet(Tcl_GetString(objv[i]));
+		    net = DefFindNet(Tcl_GetString(objv[i]));
 		    if (net == NULL) {
 			Tcl_SetResult(interp, "No such net", NULL);
 			return TCL_ERROR;
@@ -1107,7 +1127,7 @@ qrouter_stage2(ClientData clientData, Tcl_Interp *interp,
 			return TCL_ERROR;
 		    }
 		    i++;
-		    net = LookupNet(Tcl_GetString(objv[i]));
+		    net = DefFindNet(Tcl_GetString(objv[i]));
 		    if (net == NULL) {
 			Tcl_SetResult(interp, "No such net", NULL);
 			return TCL_ERROR;
@@ -1276,7 +1296,7 @@ qrouter_stage3(ClientData clientData, Tcl_Interp *interp,
 			return TCL_ERROR;
 		    }
 		    i++;
-		    net = LookupNet(Tcl_GetString(objv[i]));
+		    net = DefFindNet(Tcl_GetString(objv[i]));
 		    if (net == NULL) {
 			Tcl_SetResult(interp, "No such net", NULL);
 			return TCL_ERROR;
@@ -1362,6 +1382,74 @@ qrouter_stage3(ClientData clientData, Tcl_Interp *interp,
     return QrouterTagCallback(interp, objc, objv);
 }
 
+/*------------------------------------------------------*/
+/* Command "cleanup"					*/
+/*							*/
+/*   Clean up the nets by removing adjacent vias where	*/
+/*   such adjacent vias would cause a DRC violation.	*/
+/*   Note that this must be done between the last	*/
+/*   routing stage but before finding antenna		*/
+/*   violations, output, and delay writing, as all of	*/
+/*   these can be dependent on topology changes caused	*/
+/*   by the cleanup.					*/
+/*							*/
+/* Options:						*/
+/*							*/
+/*   cleanup all	Clean up all nets in the design	*/
+/*   cleanup net <name> [...]				*/
+/* 			Clean up the named net(s)	*/
+/*------------------------------------------------------*/
+
+static int
+qrouter_cleanup(ClientData clientData, Tcl_Interp *interp,
+               int objc, Tcl_Obj *CONST objv[])
+{
+    int result, idx, i;
+    NET net;
+
+    static char *subCmds[] = {
+	"all", "net", NULL
+    };
+    enum SubIdx {
+	AllIdx, NetIdx
+    };
+
+    if (objc < 2) {
+	Tcl_WrongNumArgs(interp, 0, objv, "?option?");
+	return TCL_ERROR;
+    }
+    else {
+	if ((result = Tcl_GetIndexFromObj(interp, objv[1],
+		(CONST84 char **)subCmds, "option", 0, &idx))
+		!= TCL_OK)
+	    return result;
+
+	// Quick check to see if cleanup_net can be avoided completely.
+	for (i = 0; i < Num_layers; i++)
+	    if (needblock[i] & (VIABLOCKX | VIABLOCKY))
+		break;
+	if (i == Num_layers) return TCL_OK;        /* No cleanup needed */
+
+	switch (idx) {
+	    case AllIdx:
+		for (i = 0; i < Numnets; i++) {
+		   net = Nlnets[i];
+		   cleanup_net(net);
+		}
+		break;
+
+	    case NetIdx:
+		for (i = 2; i < objc; i++) {
+		    net = DefFindNet(Tcl_GetString(objv[i]));
+		    if (net != NULL)
+			cleanup_net(net);
+		}
+		break;
+	}
+    }
+    return QrouterTagCallback(interp, objc, objv);
+}
+
 /*------------------------------------------------------*/
 /* Command "remove"					*/
 /*							*/
@@ -1410,7 +1498,7 @@ qrouter_remove(ClientData clientData, Tcl_Interp *interp,
 		break;
 	    case NetIdx:
 		for (i = 2; i < objc; i++) {
-		    net = LookupNet(Tcl_GetString(objv[i]));
+		    net = DefFindNet(Tcl_GetString(objv[i]));
 		    if (net != NULL)
 			ripup_net(net, (u_char)1, (u_char)1, (u_char)0);
 		}
@@ -1532,17 +1620,22 @@ qrouter_readlef(ClientData clientData, Tcl_Interp *interp,
     LEFfile = Tcl_GetString(objv[1]);
 
     mscale = LefRead(LEFfile);
-    if (Scales.mscale < mscale) Scales.mscale = mscale;
+    update_mscale(mscale);
  
     for (i = 0; i < Num_layers; i++) {
 
-       /* Set Vert and PitchX from route info */
+       /* Set Vert from route info since this gets called a lot	*/
+       /* (e.g., from eval_pt() and is more convenient to pull	*/
+       /* from an array than calling a subroutine every time.	*/
 
        Vert[i] = (1 - LefGetRouteOrientation(i));
-       PitchX[i] = LefGetRoutePitch(i);
     }
 
-    post_config();
+    /* Resolve the base horizontal and vertical pitches */
+    post_config(FALSE);
+
+    /* Set DRC blockage behavior based on via and route widths */
+    apply_drc_blocks(-1, 0.0, 0.0);
 
     return QrouterTagCallback(interp, objc, objv);
 }
@@ -1555,14 +1648,36 @@ static int
 qrouter_readdef(ClientData clientData, Tcl_Interp *interp,
                 int objc, Tcl_Obj *CONST objv[])
 {
+    char *argv;
+    u_char abort_on_error = FALSE;
+    int result;
+
+    /* Parse out options */
+
+    while (objc > 0) {
+	argv = Tcl_GetString(objv[objc - 1]);
+	if (*argv == '-') {
+	    if (!strncmp(argv + 1, "abort", 5))
+		abort_on_error = TRUE;
+	    objc--;
+	}
+	else break;
+    }
+
     if ((DEFfilename == NULL) && (objc != 2)) {
 	Tcl_SetResult(interp, "No DEF filename specified!", NULL);
 	return TCL_ERROR;
     }
+
     if (objc == 2)
-	read_def(Tcl_GetString(objv[1]));
+	result = read_def(Tcl_GetString(objv[1]));
     else
-	read_def(NULL);
+	result = read_def(NULL);
+
+    if ((result != (u_char)0) && (abort_on_error == TRUE)) {
+	Tcl_SetResult(interp, "Errors in input DEF file;  aborting.", NULL);
+	return TCL_ERROR;
+    }
 
     // Redisplay
     draw_layout();
@@ -1586,11 +1701,94 @@ qrouter_writedef(ClientData clientData, Tcl_Interp *interp,
 	Tcl_SetResult(interp, "No DEF filename specified!", NULL);
 	return TCL_ERROR;
     }
+    else DEFoutfile = DEFfilename;
 
     write_def(DEFoutfile);
     return QrouterTagCallback(interp, objc, objv);
 }
 
+/*------------------------------------------------------*/
+/* Command "antenna"					*/
+/* Use:							*/
+/*	antenna init <cellname>				*/
+/*	antenna check					*/
+/*	antenna fix					*/
+/*							*/
+/* Calculate and handle antenna violations.  Option	*/
+/* "init" declares the cellname that is an antenna	*/
+/* anchoring cell.  This must be declared before	*/
+/* routing.  "antenna check" can be called at any time	*/
+/* and reports the number of antenna violations at each	*/
+/* metal layer.  "antenna fix" attempts to fix all 	*/
+/* antenna violations by anchoring each antenna to an	*/
+/* available antenna cell tap.				*/
+/*------------------------------------------------------*/
+
+static int
+qrouter_antenna(ClientData clientData, Tcl_Interp *interp,
+                 int objc, Tcl_Obj *CONST objv[])
+{
+    char *option;
+    u_char do_fix = (u_char)0;
+
+    if (objc >= 2) {
+	option = Tcl_GetString(objv[1]);
+	if (objc == 3) antenna_cell = strdup(Tcl_GetString(objv[2]));
+
+	if (!strcmp(option, "init")) {
+	    if (objc != 3) {
+		if (antenna_cell != NULL) {
+		    Tcl_SetObjResult(interp, Tcl_NewStringObj(antenna_cell, -1));
+		}
+		else {
+		    Tcl_SetResult(interp, "No antenna cell name specified.", NULL);
+		    return TCL_ERROR;
+		}
+	    }
+	}
+	else if (!strcmp(option, "check")) {
+	    resolve_antenna(antenna_cell, (u_char)0);
+	}
+	else if (!strcmp(option, "fix")) {
+	    resolve_antenna(antenna_cell, (u_char)1);
+	}
+	else {
+	    antenna_cell = Tcl_GetString(objv[1]);
+	}
+    }
+    else {
+	Tcl_SetResult(interp, "Usage: antenna init|check|fix [cellname]", NULL);
+	return TCL_ERROR;
+    }
+
+    if (antenna_cell == NULL) {
+	Tcl_SetResult(interp, "No antenna cell specified!", NULL);
+	return TCL_ERROR;
+    }
+    return QrouterTagCallback(interp, objc, objv);
+}
+
+/*------------------------------------------------------*/
+/* Command "write_failed"				*/
+/*------------------------------------------------------*/
+
+static int
+qrouter_writefailed(ClientData clientData, Tcl_Interp *interp,
+                 int objc, Tcl_Obj *CONST objv[])
+{
+    char *outfile = NULL;
+
+    if (objc == 2)
+	outfile = Tcl_GetString(objv[1]);
+    else if (outfile == NULL) {
+	Tcl_SetResult(interp, "No output filename specified!", NULL);
+	return TCL_ERROR;
+    }
+
+    write_failed(outfile);
+    return QrouterTagCallback(interp, objc, objv);
+}
+
 /*------------------------------------------------------*/
 /* Command "write_delays"				*/
 /*------------------------------------------------------*/
@@ -1607,6 +1805,7 @@ qrouter_writedelays(ClientData clientData, Tcl_Interp *interp,
 	Tcl_SetResult(interp, "No delay filename specified!", NULL);
 	return TCL_ERROR;
     }
+    else delayoutfile = delayfilename;
 
     write_delays(delayoutfile);
     return QrouterTagCallback(interp, objc, objv);
@@ -1749,7 +1948,7 @@ qrouter_ignore(ClientData clientData, Tcl_Interp *interp,
     }
     else {
 	for (i = 1; i < objc; i++) {
-	    net = LookupNet(Tcl_GetString(objv[i]));
+	    net = DefFindNet(Tcl_GetString(objv[i]));
 	    if (net == NULL) {
 		Tcl_SetResult(interp, "No such net", NULL);
 		return TCL_ERROR;
@@ -1776,7 +1975,7 @@ static int
 qrouter_priority(ClientData clientData, Tcl_Interp *interp,
                  int objc, Tcl_Obj *CONST objv[])
 {
-    int i;
+    int i, j;
     char *netname;
     NET net;
     STRING cn, ctest;
@@ -1793,15 +1992,40 @@ qrouter_priority(ClientData clientData, Tcl_Interp *interp,
 	}
 	Tcl_SetObjResult(interp, lobj);
     }
+    else if (Nlnets == NULL) {
+	Tcl_SetResult(interp, "Must read nets from DEF file before setting priority.",
+		NULL);
+	return TCL_ERROR;
+    }
     else {
-	for (i = objc - 1; i > 0; i--) {
+
+	/* Find the highest-numbered existing critical net so that	*/
+	/* repeated calls to "priority" will work.			*/
+
+	j = -1;
+	for (i = 0; i < Numnets; i++) {
+	    net = Nlnets[i];
+	    if (net->flags & NET_CRITICAL)
+		if (net->netorder > j)
+		    j = net->netorder;
+	}
+	j++;
+
+	for (i = 1; i < objc; i++) {
 	    netname = Tcl_GetString(objv[i]);
-	    net = LookupNet(netname);
+	    net = DefFindNet(netname);
 	    if (net == NULL) {
 		Tcl_SetResult(interp, "No such net", NULL);
 	    }
-	    else {
+	    else if (!(net->flags & NET_CRITICAL)) {
 		net->flags |= NET_CRITICAL;
+		net->netorder = j++;
+
+		/* NOTE:  CriticalNet is left over from the old config	*/
+		/* file format.  Normally it will remain NULL.  If the	*/
+		/* old config file format is used, then remove items	*/
+		/* from it that match nets in the Tcl priority command.	*/
+
 		for (cn = CriticalNet; cn && cn->next; cn = cn->next) {
 		    if (!cn->next) break;
 		    if (!strcmp(cn->next->name, netname)) {
@@ -1994,11 +2218,7 @@ qrouter_layerinfo(ClientData clientData, Tcl_Interp *interp,
 /*	  a single point.  Value "none", "0", and "1"	*/
 /*	  all mean the same thing.			*/
 /*							*/
-/* via_pattern:  If vias are non-square, then they are	*/
-/*	  placed in a checkerboard pattern, with every	*/
-/*	  other via rotated 90 degrees.  If inverted,	*/
-/*	  the rotation is swapped relative to the grid	*/
-/*	  positions used in the non-inverted case.	*/
+/* via_pattern:  (deprecated)				*/
 /*							*/
 /* use: List of names of vias to use.  If any via not	*/
 /*	in this list is found when reading a .lef file	*/
@@ -2054,8 +2274,7 @@ qrouter_via(ClientData clientData, Tcl_Interp *interp,
 		    break;
 		case PatternIdx:
 		    Tcl_SetObjResult(interp,
-				Tcl_NewStringObj(
-				patternSubCmds[ViaPattern + 1], -1));
+				Tcl_NewStringObj("deprecated", -1));
 		    break;
 		case UseIdx:
 		    /* Return list of vias to use */
@@ -2097,7 +2316,6 @@ qrouter_via(ClientData clientData, Tcl_Interp *interp,
 				(CONST84 char **)patternSubCmds, "option",
 				0, &idx2)) != TCL_OK)
 			return result;
-		    ViaPattern = idx2 - 1;
 		    break;
 		case UseIdx:
 		    /* Create list of vias to use */
@@ -2114,7 +2332,7 @@ qrouter_via(ClientData clientData, Tcl_Interp *interp,
 			newVia->next = AllowedVias;
 			AllowedVias = newVia;
 		    }
-		    /* Regenerate the ViaX and ViaY lists */
+		    /* Regenerate the ViaXX, etc., lists */
 		    LefAssignLayerVias();
 		    break;
 	    }
@@ -2208,6 +2426,307 @@ qrouter_resolution(ClientData clientData, Tcl_Interp *interp,
     return QrouterTagCallback(interp, objc, objv);
 }
 
+/*------------------------------------------------------*/
+/* Command "query"					*/
+/*							*/
+/* Print information about a specific grid point,	*/
+/* instance, node, or net.  Option "watch" indicates to	*/
+/* monitor the position during DEF read-in and print	*/
+/* information when the node information changes,	*/
+/* especially if the grid position is disabled.		*/
+/*							*/
+/* Options:						*/
+/*							*/
+/*	query grid <ix> <iy> <layer> [watch]		*/
+/*	query position <dx> <dy> <layer> [watch]	*/
+/*	query instance <instance>			*/
+/*	query node <instance>/<pin>			*/
+/*	query net <name>				*/
+/*							*/
+/* <layer> may be either a layer name or integer index.	*/
+/* <dx> and <dy> should be given in microns. <ix> and	*/
+/* <iy> are integer indexes.  All other arguments are	*/
+/* strings.						*/
+/*------------------------------------------------------*/
+
+static int
+qrouter_query(ClientData clientData, Tcl_Interp *interp,
+               int objc, Tcl_Obj *CONST objv[])
+{
+    char *layername, *instname, *netname, *pinname;
+    int idx, result, layer;
+    int gridx, gridy;
+    double dx, dy;
+    unsigned char is_index, do_watch, do_unwatch;
+
+    static char *subCmds[] = {
+	"grid", "position", "instance", "node", "net", NULL
+    };
+    enum SubIdx {
+	GridIdx, PosIdx, InstIdx, NodeIdx, NetIdx
+    };
+   
+    if (objc < 2) {
+	Fprintf(stderr, "Usage:\n");
+	Fprintf(stderr, "   query position <x_microns> <y_microns> <layer> [watch]\n");
+	Fprintf(stderr, "   query grid <gridx> <gridy> <layer> [watch]\n");
+	Fprintf(stderr, "   query instance <inst_name>\n");
+	Fprintf(stderr, "   query node <inst_name>/<pin_name>\n");
+	Fprintf(stderr, "   query net <net_name>\n");
+	Tcl_WrongNumArgs(interp, 1, objv, "option ?arg?");
+	return TCL_ERROR;
+    }
+    else if ((result = Tcl_GetIndexFromObj(interp, objv[1],
+		(CONST84 char **)subCmds, "option", 0, &idx)) != TCL_OK)
+	return result;
+
+    do_watch = do_unwatch = FALSE;
+    switch (idx) {
+	case GridIdx:
+	case PosIdx:
+	    if (objc == 6) {
+		if (!strcmp(Tcl_GetString(objv[5]), "watch")) {
+		    do_watch = TRUE;
+		    objc--;
+		}
+		else if (!strcmp(Tcl_GetString(objv[5]), "unwatch")) {
+		    do_unwatch = TRUE;
+		    objc--;
+		}
+	    }
+	    if (objc != 5) {
+		Tcl_WrongNumArgs(interp, 1, objv, "option ?arg?");
+		return TCL_ERROR;
+	    }
+	    is_index = (idx == GridIdx) ? TRUE : FALSE;
+
+	    layername = Tcl_GetString(objv[4]);
+	    layer = LefFindLayerNum(layername);
+	    if (layer < 0) {
+		result = Tcl_GetIntFromObj(interp, objv[4], &layer);
+		if (result != TCL_OK) {
+		    Tcl_SetResult(interp, "No such layer name.\n", NULL);
+		    return result;
+		}
+	    }
+	    if (layer >= Num_layers) {
+		Tcl_SetResult(interp, "Bad layer number.\n", NULL);
+		return result;
+	    }
+	    if (is_index) {
+		result = Tcl_GetIntFromObj(interp, objv[2], &gridx);
+		if (result != TCL_OK) {
+		    Tcl_SetResult(interp, "Cannot parse grid X index\n", NULL);
+		    return result;
+		}
+		result = Tcl_GetIntFromObj(interp, objv[3], &gridy);
+		if (result != TCL_OK) {
+		    Tcl_SetResult(interp, "Cannot parse grid Y index\n", NULL);
+		    return result;
+		}
+		/* Translate gridx, gridy into dx, dy */
+		dx = (gridx * PitchX) + Xlowerbound;
+		dy = (gridy * PitchY) + Ylowerbound;
+	    }
+	    else {
+		result = Tcl_GetDoubleFromObj(interp, objv[2], &dx);
+		if (result != TCL_OK) {
+		    Tcl_SetResult(interp, "Cannot parse grid X position\n", NULL);
+		    return result;
+		}
+		result = Tcl_GetDoubleFromObj(interp, objv[3], &dy);
+		if (result != TCL_OK) {
+		    Tcl_SetResult(interp, "Cannot parse grid Y position\n", NULL);
+	    	    return result;
+	    	}
+	    	/* Translate dx, dy into nearest gridx, gridy */
+	    	gridx = (int)(round((dx - Xlowerbound) / PitchX));
+	    	gridy = (int)(round((dy - Ylowerbound) / PitchY));
+
+		Fprintf(stdout, "Grid position index is (%d %d)\n", gridx, gridy);
+	    }
+	    if (gridx < 0) {
+	    	Tcl_SetResult(interp, "Grid X position must not be negative.\n",
+			    NULL);
+	    	return result;
+	    }
+	    if (gridy < 0) {
+	    	Tcl_SetResult(interp, "Grid Y position must not be negative.\n",
+			    NULL);
+	    	return result;
+	    }
+	    if (do_watch) {
+		DPOINT newtest;
+		newtest = (DPOINT)malloc(sizeof(struct dpoint_));
+		newtest->layer = layer;
+		newtest->next = testpoint;
+		if (idx == GridIdx) {
+		    newtest->gridx = gridx;
+		    newtest->gridy = gridy;
+		    newtest->x = 0;
+		    newtest->y = 0;
+		    Fprintf(stdout, "Watching grid position index (%d %d)"
+			    " layer %d.\n", gridx, gridy, layer);
+		}
+		else {
+		    newtest->x = dx;
+		    newtest->y = dy;
+		    newtest->gridx = -1;
+		    newtest->gridy = -1;
+		    Fprintf(stdout, "Watching grid position (%g %g)um"
+			    " layer %d.\n", dx, dy, layer);
+		}
+		testpoint = newtest;
+	    }
+	    else if (do_unwatch) {
+		DPOINT ptest, ltest;
+		ltest = NULL;
+		for (ptest = testpoint; ptest; ptest = ptest->next) {
+		    if (ptest->x == dx && ptest->y == dy && ptest->layer == layer) {
+			if (ltest == NULL) {
+			    testpoint = testpoint->next;
+			    free(ptest);
+			}
+			else {
+			    ltest->next = ptest->next;
+			    free(ptest);
+			}
+			Fprintf(stdout, "No longer watching grid position (%g %g)um"
+				" index (%d %d) layer %d.\n",
+				dx, dy, gridx, gridy, layer);
+			break;
+		    }
+		    ltest = ptest;
+		}
+		if (ptest == NULL) {
+		    Fprintf(stdout, "Grid position (%g %g)um index (%d %d)"
+				" layer %d is not on the watch list.\n",
+				dx, dy, gridx, gridy, layer);
+		}
+	    }
+	    else if (gridx >= NumChannelsX) {
+	    	Tcl_SetResult(interp, "Grid X position is larger than the number"
+	    		" of horizontal route tracks.\n", NULL);
+	    	return result;
+	    }
+	    else if (gridy >= NumChannelsY) {
+		Tcl_SetResult(interp, "Grid Y position is larger than the number"
+			" of vertical route tracks.\n", NULL);
+		return result;
+	    }
+	    else {
+		Fprintf(stdout, "Querying grid position (%g %g)um index (%d %d)"
+			" layer %d:\n", dx, dy, gridx, gridy, layer);
+		print_grid_information(gridx, gridy, layer);
+	    }
+	    break;
+
+	case InstIdx:
+	    if (objc != 3) {
+		Tcl_WrongNumArgs(interp, 1, objv, "option ?arg?");
+		return TCL_ERROR;
+	    }
+	    instname = Tcl_GetString(objv[2]);
+	    print_instance_information(instname);
+	    break;
+
+	case NodeIdx:
+	    if (objc != 3) {
+		Tcl_WrongNumArgs(interp, 1, objv, "option ?arg?");
+		return TCL_ERROR;
+	    }
+	    instname = Tcl_GetString(objv[2]);
+	    print_node_information(instname);
+	    break;
+
+	case NetIdx:
+	    if (objc != 3) {
+		Tcl_WrongNumArgs(interp, 1, objv, "option ?arg?");
+		return TCL_ERROR;
+	    }
+	    netname = Tcl_GetString(objv[2]);
+	    print_net_information(netname);
+	    break;
+    }
+    return QrouterTagCallback(interp, objc, objv);
+}
+
+/*------------------------------------------------------*/
+/* Command "drc"					*/
+/*							*/
+/* Set qrouter options related to handling of DRC	*/
+/* violations.						*/
+/*							*/
+/* Options:						*/
+/*							*/
+/*	drc <layer>|all <dist> <dist>			*/
+/*							*/
+/* Allow exceptions to DRC handling.  Normally qrouter	*/
+/* enforces DRC distance between a via and route or	*/
+/* between two vias on adjacent tracks, forcing a 	*/
+/* keep-out area around a placed route if needed.	*/
+/* "layer" is the name of a route layer.  <dist> (value	*/
+/* in microns) will limit the enforcement if the DRC	*/
+/* violation is less than the indicated distance.	*/
+/* The first value is for via-to-via distance, and the	*/
+/* second value is for route-to-via distance.  A value	*/
+/* of zero means that the behavior is unchanged from	*/
+/* what is automatically calculated from defined	*/
+/* route and via width and spacing values.  A positive	*/
+/* distance value loosens the DRC rule, while a		*/
+/* negative distance value tightens it.			*/
+/*							*/
+/* Ignoring DRC errors is generally discouraged but	*/
+/* may be necessary in some cases if the pitch is tight	*/
+/* and assumes routes may be offset to clear vias,	*/
+/* which is something qrouter does not know how to do.	*/
+/* Only use this if routability is being impacted by	*/
+/* DRC enforcement.					*/
+/*------------------------------------------------------*/
+
+static int
+qrouter_drc(ClientData clientData, Tcl_Interp *interp,
+               int objc, Tcl_Obj *CONST objv[])
+{
+    char *layername;
+    int result, layer;
+    double routedist, viadist;
+
+    if (objc != 4) {
+	Tcl_WrongNumArgs(interp, 1, objv, "option ?arg?");
+	return TCL_ERROR;
+    }
+
+    layername = Tcl_GetString(objv[1]);
+    if (!strcasecmp(layername, "all")) {
+	layer = -1;
+    }
+    else {
+	layer = LefFindLayerNum(layername);
+	if (layer < 0) {
+	    result = Tcl_GetIntFromObj(interp, objv[1], &layer);
+	    if (result != TCL_OK) {
+		Tcl_SetResult(interp, "No such layer name.\n", NULL);
+		return result;
+	    }
+	}
+
+	if ((layer < -1) || (layer >= LefGetMaxRouteLayer())) {
+	    Tcl_SetResult(interp, "Layer number out of range.\n", NULL);
+	    return TCL_ERROR;
+	}
+    }
+
+    result = Tcl_GetDoubleFromObj(interp, objv[2], &viadist);
+    if (result != TCL_OK) return result;
+
+    result = Tcl_GetDoubleFromObj(interp, objv[3], &routedist);
+    if (result != TCL_OK) return result;
+
+    apply_drc_blocks(layer, viadist, routedist);
+
+    return QrouterTagCallback(interp, objc, objv);
+}
 
 /*------------------------------------------------------*/
 /* Command "layers"					*/
@@ -2235,10 +2754,10 @@ qrouter_layers(ClientData clientData, Tcl_Interp *interp,
     else if (objc == 2) {
 	result = Tcl_GetIntFromObj(interp, objv[1], &value);
 	if (result != TCL_OK) return result;
-	if (value <= 0 || value > LefGetMaxLayer()) {
+	if (value <= 0 || value > LefGetMaxRouteLayer()) {
 	    Tcl_SetResult(interp, "Number of layers out of range,"
 			" setting to max.", NULL);
-	    Num_layers = LefGetMaxLayer();
+	    Num_layers = LefGetMaxRouteLayer();
 	    return TCL_ERROR;
 	}
 	Num_layers = value;
@@ -2543,7 +3062,7 @@ qrouter_congested(ClientData clientData, Tcl_Interp *interp,
     else
 	entries = 0;
 
-    Congestion = (float *)calloc(NumChannelsX[0] * NumChannelsY[0],
+    Congestion = (float *)calloc(NumChannelsX * NumChannelsY,
 			sizeof(float));
 
     // Use net bounding boxes to estimate congestion
@@ -2563,8 +3082,8 @@ qrouter_congested(ClientData clientData, Tcl_Interp *interp,
 
 	for (x = net->xmin; x < net->xmax; x++)
 	    for (y = net->ymin; y < net->ymax; y++)
-		if (x >= 0 && x < NumChannelsX[0] &&
-			y >= 0 && y < NumChannelsY[0])
+		if (x >= 0 && x < NumChannelsX &&
+			y >= 0 && y < NumChannelsY)
 		    CONGEST(x, y) += density;
     }
 
@@ -2589,12 +3108,12 @@ qrouter_congested(ClientData clientData, Tcl_Interp *interp,
 	cgates[i] = (CLIST)malloc(sizeof(Clist));
 	dx = gsrch->placedX;
 	dy = gsrch->placedY;
-	bbox.x1 = (int)((dx - Xlowerbound) / PitchX[0]) - 1;
-	bbox.y1 = (int)((dy - Ylowerbound) / PitchY[0]) - 1;
+	bbox.x1 = (int)((dx - Xlowerbound) / PitchX) - 1;
+	bbox.y1 = (int)((dy - Ylowerbound) / PitchY) - 1;
 	dx = gsrch->placedX + gsrch->width;
 	dy = gsrch->placedY + gsrch->height;
-	bbox.x2 = (int)((dx - Xlowerbound) / PitchX[0]) - 1;
-	bbox.y2 = (int)((dy - Ylowerbound) / PitchY[0]) - 1;
+	bbox.x2 = (int)((dx - Xlowerbound) / PitchX) - 1;
+	bbox.y2 = (int)((dy - Ylowerbound) / PitchY) - 1;
 
 	cavg = 0.0;
 	for (x = bbox.x1; x <= bbox.x2; x++) {
@@ -2680,7 +3199,7 @@ qrouter_print(ClientData clientData, Tcl_Interp *interp,
 
     switch (idx) {
         case NetIdx:
-            net = LookupNet(Tcl_GetString(objv[2]));
+            net = DefFindNet(Tcl_GetString(objv[2]));
             if (net == NULL) {
                 Tcl_SetResult(interp, "Net not found", NULL);
                 return TCL_ERROR;
@@ -2700,7 +3219,7 @@ qrouter_print(ClientData clientData, Tcl_Interp *interp,
             break;
 
         case GateIdx:
-            gate = LookupGate(Tcl_GetString(objv[2]));
+            gate = DefFindGate(Tcl_GetString(objv[2]));
             if (gate == NULL) {
                 Tcl_SetResult(interp, "Gate not found", NULL);
                 return TCL_ERROR;
@@ -2712,3 +3231,127 @@ qrouter_print(ClientData clientData, Tcl_Interp *interp,
 }
 
 /*------------------------------------------------------*/
+/* Command "pitchx"					*/
+/*							*/
+/* Set the base pitch for vertical routing layers.	*/
+/* Values larger than current value will be ignored.	*/
+/* Without an argument, current value is returned.	*/
+/*							*/
+/* Options:						*/
+/*							*/
+/*	pitchx [<value>]				*/
+/*------------------------------------------------------*/
+
+static int
+qrouter_pitchx(ClientData clientData, Tcl_Interp *interp,
+               int objc, Tcl_Obj *CONST objv[])
+{
+    int result;
+    double value;
+
+    if (objc == 1) {
+	Tcl_SetObjResult(interp, Tcl_NewDoubleObj(PitchX));
+    }
+    else if (objc == 2) {
+	result = Tcl_GetDoubleFromObj(interp, objv[1], &value);
+	if (result != TCL_OK) return result;
+	if (value <= 0.0) {
+	    Tcl_SetResult(interp, "PitchX value has to be a positive value"
+	                  "; ignored", NULL);
+	    return TCL_ERROR;
+	}
+	else if ((PitchX > 0.0) && (value > PitchX)) {
+	    Tcl_SetResult(interp, "PitchX is larger than current value"
+	                  "; ignored", NULL);
+        }
+	else {
+	    PitchX = value;
+	}
+    }
+    else {
+	Tcl_WrongNumArgs(interp, 1, objv, "option ?arg?");
+	return TCL_ERROR;
+    }
+    return QrouterTagCallback(interp, objc, objv);
+}
+
+/*------------------------------------------------------*/
+/* Command "pitchy"					*/
+/*							*/
+/* Set the base pitch for horizontal routing layers.	*/
+/* Values larger than current value will be ignored.	*/
+/* Without an argument, current value is returned.	*/
+/*							*/
+/* Options:						*/
+/*							*/
+/*	pitchy [<value>]				*/
+/*------------------------------------------------------*/
+
+static int
+qrouter_pitchy(ClientData clientData, Tcl_Interp *interp,
+               int objc, Tcl_Obj *CONST objv[])
+{
+    int result;
+    double value;
+
+    if (objc == 1) {
+	Tcl_SetObjResult(interp, Tcl_NewDoubleObj(PitchY));
+    }
+    else if (objc == 2) {
+	result = Tcl_GetDoubleFromObj(interp, objv[1], &value);
+	if (result != TCL_OK) return result;
+	if (value <= 0.0) {
+	    Tcl_SetResult(interp, "PitchY value has to be a positive value"
+	                  "; ignored", NULL);
+	    return TCL_ERROR;
+	}
+	else if ((PitchY > 0.0) && (value > PitchY)) {
+	    Tcl_SetResult(interp, "PitchY is larger than current value"
+	                  "; ignored", NULL);
+        }
+	else {
+	    PitchY = value;
+	}
+    }
+    else {
+	Tcl_WrongNumArgs(interp, 1, objv, "option ?arg?");
+	return TCL_ERROR;
+    }
+    return QrouterTagCallback(interp, objc, objv);
+}
+
+/*------------------------------------------------------*/
+/* Command "unblock"					*/
+/*							*/
+/* Set the unblocking flag, which indicates to qrouter	*/
+/* that all grid points that lie cleanly inside pin	*/
+/* geometry should be marked routable.  This command	*/
+/* must be issued before read_def to be effective.	*/
+/*							*/
+/* Options:						*/
+/*							*/
+/*	unblock	[true|false]				*/
+/*------------------------------------------------------*/
+
+static int
+qrouter_unblock(ClientData clientData, Tcl_Interp *interp,
+               int objc, Tcl_Obj *CONST objv[])
+{
+    int result, value;
+
+    if (objc == 1) {
+	Tcl_SetObjResult(interp, Tcl_NewBooleanObj(unblockAll));
+    }
+    else if (objc == 2) {
+	result = Tcl_GetBooleanFromObj(interp, objv[1], &value);
+	if (result != TCL_OK) return result;
+	unblockAll = (value == 0) ? FALSE : TRUE;
+    }
+    else {
+	Tcl_WrongNumArgs(interp, 1, objv, "option ?arg?");
+	return TCL_ERROR;
+    }
+    return QrouterTagCallback(interp, objc, objv);
+}
+
+/*------------------------------------------------------*/
